import re

from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import (
    QAction,
    QCursor,
    QFont,
)
from PySide6.QtWidgets import (
    QFontDialog,
    QMenu,
)

from plover import _
from plover.suggestions import Suggestion
from plover.formatting import RetroFormatter

from plover.gui_qt import utils
from plover.gui_qt.suggestions_dialog_ui import Ui_SuggestionsDialog
from plover.gui_qt.utils import ToolBar
from plover.gui_qt.tool import Tool


class SuggestionsDialog(Tool, Ui_SuggestionsDialog):
    # i18n: Widget: “SuggestionsDialog”, tooltip.
    __doc__ = _("Suggest possible strokes for the last written words.")

    TITLE = _("Suggestions")
    ICON = ":/resources/lightbulb.svg"
    ROLE = "suggestions"
    SHORTCUT = "Ctrl+J"

    WORD_RX = re.compile(r"(?:\w+|[^\w\s]+)\s*")

    STYLE_TRANSLATION, STYLE_STROKES = range(2)

    # Anatomy of the text document:
    # - "root" frame:
    #  - 0+ "suggestions" frames
    #   - 1+ "translation" frames
    #    - 1-10 "strokes" frames

    def __init__(self, engine):
        super().__init__(engine)
        self.setupUi(self)
        self._last_suggestions = None
        # Toolbar.
        actions = [
            self.action_SelectFont,
            self.action_Clear,
        ]
        if not utils.is_wayland():
            # Wayland does not support window on top.
            actions.insert(0, self.action_ToggleOnTop)

        self.layout().addWidget(ToolBar(*actions))

        self.action_Clear.setEnabled(False)
        # Font popup menu.
        self._font_menu = QMenu()
        # i18n: Widget: “SuggestionsDialog”, “font” menu.
        self._font_menu_text = QAction(_("&Text"), self._font_menu)
        # i18n: Widget: “SuggestionsDialog”, “font” menu.
        self._font_menu_strokes = QAction(_("&Strokes"), self._font_menu)
        self._font_menu.addActions([self._font_menu_text, self._font_menu_strokes])
        engine.signal_connect("translated", self.on_translation)
        self.suggestions.setFocus()
        self.restore_state()
        self.finished.connect(self.save_state)

    def _get_font(self, name):
        return getattr(self.suggestions, name)

    def _set_font(self, name, font):
        setattr(self.suggestions, name, font)

    def _restore_state(self, settings):
        for name in (
            "text_font",
            "strokes_font",
        ):
            font_string = settings.value(name)
            if font_string is None:
                continue
            font = QFont()
            if not font.fromString(font_string):
                continue
            self._set_font(name, font)
        ontop = settings.value("ontop", None, bool)
        if ontop is not None:
            self.action_ToggleOnTop.setChecked(ontop)
            self.toggle_ontop(ontop)

    def _save_state(self, settings):
        for name in (
            "text_font",
            "strokes_font",
        ):
            font = self._get_font(name)
            font_string = font.toString()
            settings.setValue(name, font_string)
        ontop = bool(self.windowFlags() & Qt.WindowType.WindowStaysOnTopHint)
        settings.setValue("ontop", ontop)

    def _show_suggestions(self, suggestion_list):
        self.suggestions.append(suggestion_list)
        self.action_Clear.setEnabled(True)

    @staticmethod
    def tails(ls):
        """Return all tail combinations (a la Haskell)

        tails :: [x] -> [[x]]
        >>> tails('abcd')
        ['abcd', 'bcd', 'cd', d']

        """

        for i in range(len(ls)):
            yield ls[i:]

    def on_translation(self, old, new):
        # Check for new output.
        for a in reversed(new):
            if a.text and not a.text.isspace():
                break
        else:
            return

        # Get the last 10 words.
        with self._engine:
            last_translations = self._engine.translator_state.translations
            retro_formatter = RetroFormatter(last_translations)
            split_words = retro_formatter.last_words(10, rx=self.WORD_RX)

        suggestion_list = []
        for phrase in self.tails(split_words):
            phrase = "".join(phrase)
            suggestion_list.extend(self._engine.get_suggestions(phrase))

        if not suggestion_list and split_words:
            suggestion_list = [Suggestion(split_words[-1], [])]

        if suggestion_list and suggestion_list != self._last_suggestions:
            self._last_suggestions = suggestion_list
            self._show_suggestions(suggestion_list)

    @Slot()
    def select_font(self):
        action = self._font_menu.exec(QCursor.pos())
        if action is None:
            return
        if action == self._font_menu_text:
            name = "text_font"
        elif action == self._font_menu_strokes:
            name = "strokes_font"
        font = self._get_font(name)
        ok, font = QFontDialog.getFont(font, self, "")
        if ok:
            self._set_font(name, font)

    @Slot(bool)
    def toggle_ontop(self, ontop):
        self.setWindowFlag(Qt.WindowStaysOnTopHint, ontop)
        self.show()

    @Slot()
    def clear(self):
        self.action_Clear.setEnabled(False)
        self._last_suggestions = None
        self.suggestions.clear()



from time import sleep

from plover.output import Output


class GenericKeyboardEmulation(Output):
    def __init__(self):
        super().__init__()
        self._key_press_delay_ms = 0

    def set_key_press_delay(self, delay_ms):
        self._key_press_delay_ms = delay_ms

    def delay(self):
        if self._key_press_delay_ms > 0:
            sleep(self._key_press_delay_ms / 1000)

    def half_delay(self):
        if self._key_press_delay_ms > 0:
            sleep(self._key_press_delay_ms / 2000)

    def with_delay(self, iterable):
        for item in iterable:
            yield item
            self.delay()



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for config.py."""

from ast import literal_eval
from contextlib import ExitStack
from pathlib import Path
from site import USER_BASE
from string import Template
import inspect
import json
import os
import subprocess
import sys
import textwrap

import appdirs
import pytest

from plover import config
from plover.config import DictionaryConfig
from plover.oslayer.config import PLATFORM
from plover.machine.keyboard import Keyboard
from plover.machine.keymap import Keymap
from plover.misc import expand_path
from plover.registry import Registry
from plover.system import english_stenotype


def dedent_strip(s):
    return textwrap.dedent(s).strip()


def dict_replace(d, update):
    d = dict(d)
    d.update(update)
    return d


class FakeMachine:
    KEYMAP_MACHINE_TYPE = "Keyboard"

    def get_keys():
        return Keyboard.get_keys()

    def get_actions():
        return Keyboard.get_actions()

    @staticmethod
    def get_option_info():
        bool_converter = lambda s: s == "True"
        return {
            "stroption1": (None, str),
            "intoption1": (3, int),
            "stroption2": ("abc", str),
            "floatoption1": (1, float),
            "booloption1": (True, bool_converter),
            "booloption2": (False, bool_converter),
        }


FakeMachine.DEFAULT_OPTIONS = {
    k: v[0] for k, v in FakeMachine.get_option_info().items()
}


class FakeSystem:
    KEYS = english_stenotype.KEYS
    IMPLICIT_HYPHEN_KEYS = ()
    SUFFIX_KEYS = ()
    NUMBER_KEY = english_stenotype.NUMBER_KEY
    NUMBERS = english_stenotype.NUMBERS
    UNDO_STROKE_STENO = english_stenotype.UNDO_STROKE_STENO
    ORTHOGRAPHY_RULES = []
    ORTHOGRAPHY_RULES_ALIASES = {}
    ORTHOGRAPHY_WORDLIST = None
    KEYMAPS = {
        "Faky faky": english_stenotype.KEYMAPS["Keyboard"],
    }
    DEFAULT_DICTIONARIES = ("utilisateur.json", "principal.json")


def test_config_dict():
    short_path = os.path.normcase(os.path.normpath("~/foo/bar"))
    full_path = os.path.normcase(os.path.expanduser(os.path.normpath("~/foo/bar")))
    # Path should be expanded.
    assert DictionaryConfig(short_path).path == full_path
    assert DictionaryConfig(full_path).path == full_path
    # Short path is available through `short_path`.
    assert DictionaryConfig(full_path).short_path == short_path
    assert DictionaryConfig(short_path).short_path == short_path
    # Enabled default to True.
    assert DictionaryConfig("foo").enabled
    assert not DictionaryConfig("foo", False).enabled
    # When converting to a dict (for dumping to JSON),
    # a dictionary with the shortened path is used.
    assert DictionaryConfig(full_path).to_dict() == {
        "path": short_path,
        "enabled": True,
    }
    assert DictionaryConfig(short_path, False).to_dict() == {
        "path": short_path,
        "enabled": False,
    }
    # Test from_dict creation helper.
    assert DictionaryConfig.from_dict({"path": short_path}) == DictionaryConfig(
        short_path
    )
    assert DictionaryConfig.from_dict(
        {"path": full_path, "enabled": False}
    ) == DictionaryConfig(short_path, False)


if PLATFORM == "win":
    ABS_PATH = os.path.normcase(r"c:/foo/bar")
else:
    ABS_PATH = "/foo/bar"

DEFAULT_KEYMAP = Keymap(
    Keyboard.get_keys(), english_stenotype.KEYS + Keyboard.get_actions()
)
DEFAULT_KEYMAP.set_mappings(english_stenotype.KEYMAPS["Keyboard"])

DEFAULTS = {
    "space_placement": "Before Output",
    "start_attached": False,
    "start_capitalized": False,
    "undo_levels": config.DEFAULT_UNDO_LEVELS,
    "log_file_name": expand_path("strokes.log"),
    "enable_stroke_logging": False,
    "enable_translation_logging": False,
    "start_minimized": False,
    "show_stroke_display": False,
    "show_suggestions_display": False,
    "translation_frame_opacity": 100,
    "classic_dictionaries_display_order": False,
    "enabled_extensions": set(),
    "auto_start": False,
    "machine_type": "Keyboard",
    "machine_specific_options": {"arpeggiate": False, "first_up_chord_send": False},
    "system_name": config.DEFAULT_SYSTEM_NAME,
    "system_keymap": DEFAULT_KEYMAP,
    "dictionaries": [
        DictionaryConfig(p) for p in english_stenotype.DEFAULT_DICTIONARIES
    ],
}

CONFIG_TESTS = (
    (
        "defaults",
        """
     """,
        DEFAULTS,
        {},
        {},
        """
     """,
    ),
    (
        "simple_options",
        """
     [Output Configuration]
     space_placement = After Output
     start_attached = true
     start_capitalized = yes
     undo_levels = 42
     """,
        dict_replace(
            DEFAULTS,
            {
                "space_placement": "After Output",
                "start_attached": True,
                "start_capitalized": True,
                "undo_levels": 42,
            },
        ),
        {
            "space_placement": "Before Output",
            "start_attached": False,
            "start_capitalized": False,
            "undo_levels": 200,
        },
        None,
        """
     [Output Configuration]
     space_placement = Before Output
     start_attached = False
     start_capitalized = False
     undo_levels = 200
     """,
    ),
    (
        "machine_options",
        """
     [Machine Configuration]
     auto_start = True
     machine_type = keyboard

     [Keyboard]
     arpeggiate = True
     """,
        dict_replace(
            DEFAULTS,
            {
                "auto_start": True,
                "machine_specific_options": dict_replace(
                    DEFAULTS["machine_specific_options"],
                    {
                        "arpeggiate": True,
                    },
                ),
            },
        ),
        {
            "machine_type": "faKY FAky",
            "machine_specific_options": {
                "stroption1": 42,
                "floatoption1": "4.2",
                "booloption1": False,
            },
            "system_name": "FAUX SYSTÈME",
            "system_keymap": str(DEFAULT_KEYMAP),
        },
        {
            "machine_type": "Faky faky",
            "machine_specific_options": dict_replace(
                FakeMachine.DEFAULT_OPTIONS,
                {
                    "stroption1": "42",
                    "floatoption1": 4.2,
                    "booloption1": False,
                },
            ),
            "system_name": "Faux système",
            "system_keymap": DEFAULT_KEYMAP,
            "dictionaries": [
                DictionaryConfig("utilisateur.json"),
                DictionaryConfig("principal.json"),
            ],
        },
        """
     [Machine Configuration]
     auto_start = True
     machine_type = Faky faky

     [Keyboard]
     arpeggiate = True

     [Faky faky]
     booloption1 = False
     booloption2 = False
     floatoption1 = 4.2
     intoption1 = 3
     stroption1 = 42
     stroption2 = abc

     [System]
     name = Faux système

     [System: Faux système]
     keymap[faky faky] = %s
     """
        % DEFAULT_KEYMAP,
    ),
    (
        "machine_bool_option",
        """
     """,
        DEFAULTS,
        {
            "machine_specific_options": {
                "arpeggiate": True,
            },
        },
        {
            "machine_specific_options": {
                "arpeggiate": True,
                "first_up_chord_send": False,
            }
        },
        """
     [Keyboard]
     arpeggiate = True
     first_up_chord_send = False
     """,
    ),
    (
        "legacy_dictionaries_1",
        """
     [Dictionary Configuration]
     dictionary_file = main.json
     """,
        dict_replace(
            DEFAULTS,
            {
                "dictionaries": [DictionaryConfig("main.json")],
            },
        ),
        {
            "dictionaries": ["user.json", "main.json"],
        },
        {
            "dictionaries": [
                DictionaryConfig("user.json"),
                DictionaryConfig("main.json"),
            ],
        },
        """
     [System: English Stenotype]
     dictionaries = [{"enabled": true, "path": "user.json"}, {"enabled": true, "path": "main.json"}]
     """,
    ),
    (
        "legacy_dictionaries_2",
        """
     [Dictionary Configuration]
     dictionary_file1 = main.json
     dictionary_file3 = user.json
     """,
        dict_replace(
            DEFAULTS,
            {
                "dictionaries": [
                    DictionaryConfig("user.json"),
                    DictionaryConfig("main.json"),
                ],
            },
        ),
        {
            "dictionaries": ["user.json", "commands.json", "main.json"],
        },
        {
            "dictionaries": [
                DictionaryConfig("user.json"),
                DictionaryConfig("commands.json"),
                DictionaryConfig("main.json"),
            ],
        },
        """
     [System: English Stenotype]
     dictionaries = [{"enabled": true, "path": "user.json"}, {"enabled": true, "path": "commands.json"}, {"enabled": true, "path": "main.json"}]
     """,
    ),
    (
        "dictionaries",
        """
     [System: English Stenotype]
     dictionaries = %s
     """
        % json.dumps([os.path.join(ABS_PATH, "user.json"), "english/main.json"]),
        dict_replace(
            DEFAULTS,
            {
                "dictionaries": [
                    DictionaryConfig(os.path.join(ABS_PATH, "user.json")),
                    DictionaryConfig("english/main.json"),
                ],
            },
        ),
        {
            "dictionaries": [
                DictionaryConfig(os.path.join(ABS_PATH, "user.json")),
                DictionaryConfig("english/main.json"),
            ],
        },
        {
            "dictionaries": [
                DictionaryConfig(os.path.join(ABS_PATH, "user.json")),
                DictionaryConfig("english/main.json"),
            ],
        },
        """
     [System: English Stenotype]
     dictionaries = %s
     """
        % json.dumps(
            [
                {"enabled": True, "path": os.path.join(ABS_PATH, "user.json")},
                {"enabled": True, "path": os.path.join("english", "main.json")},
            ],
            sort_keys=True,
        ),
    ),
    (
        "invalid_config",
        """
     [Startup]
     Start Minimized = True

     [Machine Configuratio
     machine_type = Faky faky
     """,
        config.InvalidConfigurationError,
        {},
        {},
        """
     """,
    ),
    (
        "invalid_keymap_1",
        """
     [System: English Stenotype]
     keymap[keyboard] = [["_-"]]
     """,
        DEFAULTS,
        {},
        {},
        None,
    ),
    (
        "invalid_keymap_2",
        """
     [System: English Stenotype]
     keymap[keyboard] = 42
     """,
        DEFAULTS,
        {},
        {},
        None,
    ),
    (
        "invalid_options_1",
        """
     [System]
     name = foobar
     """,
        DEFAULTS,
        {},
        {},
        None,
    ),
    (
        "invalid_options_2",
        """
     [Machine Configuration]
     machine_type = Faky faky

     [Faky faky]
     booloption2 = 50
     intoption1 = 3.14
     """,
        dict_replace(
            DEFAULTS,
            {
                "machine_type": "Faky faky",
                "machine_specific_options": FakeMachine.DEFAULT_OPTIONS,
            },
        ),
        {},
        {},
        None,
    ),
    (
        "invalid_options_3",
        """
     [Output Configuration]
     undo_levels = foobar
     """,
        DEFAULTS,
        {},
        {},
        None,
    ),
    (
        "invalid_update_1",
        """
     [Translation Frame]
     opacity = 75
     """,
        dict_replace(
            DEFAULTS,
            {
                "translation_frame_opacity": 75,
            },
        ),
        {
            "start_minimized": False,
            "show_stroke_display": True,
            "translation_frame_opacity": 101,
        },
        config.InvalidConfigOption,
        None,
    ),
    (
        "invalid_update_2",
        """
     """,
        DEFAULTS,
        {
            "machine_type": "FakeMachine",
            "machine_specific_options": dict_replace(
                FakeMachine.DEFAULT_OPTIONS,
                {
                    "stroption1": "42",
                },
            ),
        },
        config.InvalidConfigOption,
        None,
    ),
    (
        "setitem_valid",
        """
     """,
        DEFAULTS,
        ("auto_start", True),
        None,
        """
     [Machine Configuration]
     auto_start = True
     """,
    ),
    (
        "setitem_invalid",
        """
     """,
        DEFAULTS,
        ("undo_levels", -42),
        config.InvalidConfigOption,
        """
     """,
    ),
)


@pytest.mark.parametrize(
    (
        "original_contents",
        "original_config",
        "config_update",
        "validated_config_update",
        "resulting_contents",
    ),
    [t[1:] for t in CONFIG_TESTS],
    ids=[t[0] for t in CONFIG_TESTS],
)
def test_config(
    original_contents,
    original_config,
    config_update,
    validated_config_update,
    resulting_contents,
    monkeypatch,
    tmpdir,
):
    registry = Registry()
    registry.register_plugin("machine", "Keyboard", Keyboard)
    registry.register_plugin("machine", "Faky faky", FakeMachine)
    registry.register_plugin("system", "English Stenotype", english_stenotype)
    registry.register_plugin("system", "Faux système", FakeSystem)
    monkeypatch.setattr("plover.config.registry", registry)
    config_file = tmpdir / "config.cfg"
    # Check initial contents.
    config_file.write_text(original_contents, encoding="utf-8")
    cfg = config.Config(config_file.strpath)
    if inspect.isclass(original_config):
        with pytest.raises(original_config):
            cfg.load()
        original_config = dict(DEFAULTS)
        cfg.clear()
    else:
        cfg.load()
    cfg_dict = cfg.as_dict()
    for name, value in original_config.items():
        assert cfg[name] == value
        assert cfg_dict[name] == value
    # Check updated contents.
    with ExitStack() as stack:
        if inspect.isclass(validated_config_update):
            stack.enter_context(pytest.raises(validated_config_update))
            validated_config_update = None
        elif validated_config_update is None:
            validated_config_update = config_update
        if isinstance(config_update, dict):
            cfg.update(**config_update)
        else:
            key, value = config_update
            cfg[key] = value
        if validated_config_update is not None:
            if isinstance(validated_config_update, dict):
                cfg_dict.update(validated_config_update)
            else:
                key, value = validated_config_update
                cfg_dict[key] = value
        assert cfg.as_dict() == cfg_dict
    config_file.write_text("", encoding="utf-8")
    cfg.save()
    if resulting_contents is None:
        resulting_contents = original_contents
    assert config_file.read_text(encoding="utf-8").strip() == dedent_strip(
        resulting_contents
    )


CONFIG_MISSING_INTS_TESTS = (
    (
        "int_option",
        config.OUTPUT_CONFIG_SECTION,
        "undo_levels",
        config.DEFAULT_UNDO_LEVELS,
    ),
    (
        "opacity_option",
        "Translation Frame",
        "translation_frame_opacity",
        100,
    ),
)


@pytest.mark.parametrize(
    ("which_section", "which_option", "fixed_value"),
    [t[1:] for t in CONFIG_MISSING_INTS_TESTS],
    ids=[t[0] for t in CONFIG_MISSING_INTS_TESTS],
)
def test_config_missing_ints(
    which_section, which_option, fixed_value, monkeypatch, tmpdir, caplog
):
    registry = Registry()
    registry.register_plugin("machine", "Keyboard", Keyboard)
    registry.register_plugin("system", "English Stenotype", english_stenotype)
    monkeypatch.setattr("plover.config.registry", registry)
    config_file = tmpdir / "config.cfg"

    # Make config with the appropriate empty section
    contents = f"""
    [{which_section}]
    """
    config_file.write_text(contents, encoding="utf-8")
    cfg = config.Config(config_file.strpath)
    cfg.load()

    # Try to access an option under that section
    # (should trigger validation)
    assert cfg[which_option] == fixed_value

    # Ensure that missing options are handled
    assert "InvalidConfigOption: None" not in caplog.text
    # ... or any that there aren't any unhandled errors
    for record in caplog.records:
        assert record.levelname != "ERROR"


CONFIG_DIR_TESTS = (
    # Default to `user_config_dir`.
    (
        """
     $user_config_dir/foo.cfg
     $user_data_dir/bar.cfg
     $cwd/config.cfg
     """,
        "$user_config_dir",
    ),
    # Use cwd if config file is present, override other directories.
    (
        """
     $user_config_dir/plover.cfg
     $user_data_dir/plover.cfg
     $cwd/plover.cfg
     """,
        "$cwd",
    ),
    # plover.cfg must be a file.
    (
        """
     $user_data_dir/plover.cfg
     $cwd/plover.cfg/config
     """,
        "$user_data_dir",
    ),
)

if appdirs.user_data_dir() != appdirs.user_config_dir():
    CONFIG_DIR_TESTS += (
        # `user_config_dir` take precedence over `user_data_dir`.
        (
            """
         $user_config_dir/plover.cfg
         $user_data_dir/plover.cfg
         $cwd/config.cfg
         """,
            "$user_config_dir",
        ),
        # But `user_data_dir` is still used when applicable.
        (
            """
         $user_config_dir/plover.cfg/conf
         $user_data_dir/plover.cfg
         $cwd/config.cfg
         """,
            "$user_data_dir",
        ),
    )


@pytest.mark.parametrize(("tree", "expected_config_dir"), CONFIG_DIR_TESTS)
def test_config_dir(tree, expected_config_dir, tmpdir):
    # Create fake home/cwd directories.
    home = tmpdir / "home"
    home.mkdir()
    cwd = tmpdir / "cwd"
    cwd.mkdir()
    directories = {
        "tmpdir": str(tmpdir),
        "home": str(home),
        "cwd": str(cwd),
    }
    # Setup environment.
    env = dict(os.environ)
    if PLATFORM == "win":
        env["USERPROFILE"] = str(home)
    else:
        env["HOME"] = str(home)
    env["PYTHONUSERBASE"] = USER_BASE
    # Ensure XDG_xxx environment variables don't screw up our isolation.
    for k in list(env):
        if k.startswith("XDG_"):
            del env[k]

    # Helpers.
    def pyeval(script):
        return literal_eval(
            subprocess.check_output(
                (sys.executable, "-c", script), cwd=str(cwd), env=env
            ).decode()
        )

    def path_expand(path):
        return Template(path.replace("/", os.sep)).substitute(**directories)

    # Find out user_config_dir/user_data_dir locations.
    directories.update(
        pyeval(
            dedent_strip(
                """
        import appdirs, os
        print(repr({
            'user_config_dir': appdirs.user_config_dir('plover'),
            'user_data_dir': appdirs.user_data_dir('plover'),
        }))
        """
            )
        )
    )
    # Create initial tree.
    created_files = []
    for filename_template in dedent_strip(tree).replace("/", os.sep).split("\n"):
        filename = Path(path_expand(filename_template))
        filename.parent.mkdir(parents=True, exist_ok=True)
        filename.write_text("pouet")
        created_files.append(filename)
    # Check plover.oslayer.config.CONFIG_DIR is correctly set.
    config_dir = pyeval(
        dedent_strip(
            """
        __import__('sys').path.insert(0, %r)
        from plover.oslayer.config import CONFIG_DIR
        print(repr(CONFIG_DIR))
        """
            % str(Path(config.__file__).parent.parent)
        )
    )
    expected_config_dir = path_expand(expected_config_dir)
    assert config_dir == expected_config_dir
    # Cleanup created files and directories.
    for file in created_files:
        try:
            file.unlink()  # Delete the file
        except FileNotFoundError:
            pass
    for file in created_files:
        try:
            file.parent.rmdir()  # Attempt to remove empty directories
        except OSError:
            pass



from collections import defaultdict

from pkginfo.distribution import Distribution as Metadata
from importlib.metadata import distributions

from plover.plugins_manager.plugin_metadata import PluginMetadata
from plover import log


def list_plugins():
    plugins = defaultdict(list)

    # Iterate over all installed distributions
    for dist in distributions():
        if dist.metadata["Name"].lower() == "plover":
            continue

        # Check if any entry point group starts with 'plover.'
        if not any(ep.group.startswith("plover.") for ep in dist.entry_points):
            continue

        # Determine the metadata entry type
        metadata_entry = "METADATA" if dist.files else "PKG-INFO"

        # Check if the distribution has metadata
        try:
            metadata = Metadata()
            metadata.parse(dist.read_text(metadata_entry))
        except (KeyError, FileNotFoundError):
            log.warning(
                "ignoring distribution (missing metadata): %s", dist.metadata["Name"]
            )
            continue

        # Create PluginMetadata from the parsed metadata
        plugin_metadata = PluginMetadata.from_dict(
            {attr: getattr(metadata, attr, "") for attr in PluginMetadata._fields}
        )
        plugins[dist.metadata["Name"].lower()].append(plugin_metadata)

    # Sort and return plugins
    return {name: list(sorted(versions)) for name, versions in plugins.items()}



from PySide6.QtWidgets import QDialogButtonBox

from plover import _

from plover.gui_qt.add_translation_dialog_ui import Ui_AddTranslationDialog
from plover.gui_qt.tool import Tool


class AddTranslationDialog(Tool, Ui_AddTranslationDialog):
    # i18n: Widget: “AddTranslationDialog”, tooltip.
    __doc__ = _("Add a new translation to the dictionary.")

    TITLE = _("Add Translation")
    ICON = ":/resources/translation_add.svg"
    ROLE = "add_translation"
    SHORTCUT = "Ctrl+N"

    def __init__(self, engine, dictionary_path=None):
        super().__init__(engine)
        self.setupUi(self)
        self.add_translation.select_dictionary(dictionary_path)
        self.add_translation.mappingValid.connect(self.on_mapping_valid)
        self.on_mapping_valid(self.add_translation.mapping_is_valid)
        engine.signal_connect("config_changed", self.on_config_changed)
        self.on_config_changed(engine.config)
        self.installEventFilter(self)
        self.restore_state()
        self.finished.connect(self.save_state)

    def on_mapping_valid(self, valid):
        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(valid)

    def on_config_changed(self, config_update):
        if "translation_frame_opacity" in config_update:
            opacity = config_update.get("translation_frame_opacity")
            if opacity is None:
                return
            assert 0 <= opacity <= 100
            self.setWindowOpacity(opacity / 100.0)

    def accept(self):
        self.add_translation.save_entry()
        super().accept()

    def reject(self):
        self.add_translation.reject()
        super().reject()



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Generic stenography data models.

This module contains the following class:

Stroke -- A data model class that encapsulates a sequence of steno keys.

"""

from plover_stroke import BaseStroke


class Stroke(BaseStroke):
    """A standardized data model for stenotype machine strokes.

    This class standardizes the representation of a stenotype chord. A stenotype
    chord can be any sequence of stenotype keys that can be simultaneously
    pressed. Nearly all stenotype machines offer the same set of keys that can
    be combined into a chord, though some variation exists due to duplicate
    keys. This class accounts for such duplication, imposes the standard
    stenographic ordering on the keys, and combines the keys into a single
    string (called RTFCRE for historical reasons).

    """

    PREFIX_STROKE = None
    UNDO_STROKE = None

    @classmethod
    def setup(
        cls,
        keys,
        implicit_hyphen_keys,
        number_key,
        numbers,
        feral_number_key,
        undo_stroke,
    ):
        if number_key is None:
            assert not numbers
            numbers = None
        super().setup(keys, implicit_hyphen_keys, number_key, numbers, feral_number_key)
        cls._class = type(cls.__name__, (cls,), {"_helper": cls._helper})
        cls._class._class = cls._class
        cls._class.PREFIX_STROKE = cls.PREFIX_STROKE = cls.from_integer(0)
        cls._class.UNDO_STROKE = cls.UNDO_STROKE = cls.from_steno(undo_stroke)

    @classmethod
    def from_steno(cls, steno):
        return int.__new__(cls._class, cls._helper.stroke_from_steno(steno))

    @classmethod
    def from_keys(cls, keys):
        return int.__new__(cls._class, cls._helper.stroke_from_keys(keys))

    @classmethod
    def from_integer(cls, integer):
        return int.__new__(cls._class, cls._helper.stroke_from_int(integer))

    @classmethod
    def normalize_stroke(cls, steno, strict=True):
        try:
            return cls._helper.normalize_stroke(steno)
        except ValueError:
            if strict:
                raise
            return steno

    @classmethod
    def normalize_steno(cls, steno, strict=True):
        try:
            return cls._helper.normalize_steno(steno)
        except ValueError:
            if strict:
                raise
            return tuple(steno.split("/"))

    @classmethod
    def steno_to_sort_key(cls, steno, strict=True):
        try:
            return cls._helper.steno_to_sort_key(steno)
        except ValueError:
            if strict:
                raise
            return b"\x00\x00" + steno.encode("utf-8")

    def __new__(cls, value):
        return int.__new__(cls._class, cls._helper.stroke_from_any(value))

    @property
    def steno_keys(self):
        return list(self.keys())

    @property
    def rtfcre(self):
        return self._helper.stroke_to_steno(self)

    @property
    def is_correction(self):
        return int(self) == int(self.UNDO_STROKE)

    def __str__(self):
        prefix = "*" if self.is_correction else ""
        return f"{prefix}Stroke({self.rtfcre} : {self.steno_keys})"

    __repr__ = __str__


normalize_stroke = Stroke.normalize_stroke
normalize_steno = Stroke.normalize_steno
steno_to_sort_key = Stroke.steno_to_sort_key


def sort_steno_strokes(strokes_list):
    """Return suggestions, sorted by fewest strokes, then fewest keys."""
    return sorted(strokes_list, key=lambda x: (len(x), sum(map(len, x))))



import inspect

import pytest


def parametrize(tests, arity=None):
    """Helper for parametrizing pytest tests.

    Expects a list of lambdas, one per test. Each lambda must return
    the parameters for its respective test.

    Test identifiers will be automatically generated, from the test
    number and its lambda definition line (1.10, 2.12, 3.20, ...).

    If arity is None, the arguments being parametrized will be automatically
    set from the function's last arguments, according to the numbers of
    parameters for each test.

    """
    ids = []
    argvalues = []
    for n, t in enumerate(tests):
        line = inspect.getsourcelines(t)[1]
        ids.append("%u:%u" % (n + 1, line))
        argvalues.append(t())
    if arity is None:
        arity = len(argvalues[0])
    assert arity > 0

    def decorator(fn):
        argnames = list(
            parameter.name
            for parameter in inspect.signature(fn).parameters.values()
            if parameter.default is inspect.Parameter.empty
        )[-arity:]
        if arity == 1:
            argnames = argnames[0]
        return pytest.mark.parametrize(argnames, argvalues, ids=ids)(fn)

    return decorator



from Cocoa import (
    NSWorkspace,
    NSRunningApplication,
    NSApplicationActivateIgnoringOtherApps,
)


def GetForegroundWindow():
    return NSWorkspace.sharedWorkspace().frontmostApplication().processIdentifier()


def SetForegroundWindow(pid):
    target_window = NSRunningApplication.runningApplicationWithProcessIdentifier_(pid)
    target_window.activateWithOptions_(NSApplicationActivateIgnoringOtherApps)



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Parsing a json formatted dictionary."""

try:
    import simplejson as json
except ImportError:
    import json

from plover.dictionary.helpers import StenoNormalizer
from plover.steno_dictionary import StenoDictionary
from plover.steno import steno_to_sort_key


class JsonDictionary(StenoDictionary):
    def _load(self, filename):
        with open(filename, "rb") as fp:
            contents = fp.read()
        for encoding in ("utf-8", "latin-1"):
            try:
                contents = contents.decode(encoding)
            except UnicodeDecodeError:
                continue
            else:
                break
        else:
            raise ValueError("'%s' encoding could not be determined" % (filename,))
        d = dict(json.loads(contents))
        with StenoNormalizer(filename) as normalize_steno:
            self.update((normalize_steno(x[0]), x[1]) for x in d.items())

    def _save(self, filename):
        mappings = [("/".join(k), v) for k, v in self.items()]
        mappings.sort(key=lambda i: steno_to_sort_key(i[0], strict=False))
        with open(filename, "w", encoding="utf-8", newline="\n") as fp:
            json.dump(
                dict(mappings), fp, ensure_ascii=False, indent=0, separators=(",", ": ")
            )
            fp.write("\n")



from ctypes import windll, wintypes


GetForegroundWindow = windll.user32.GetForegroundWindow
GetForegroundWindow.argtypes = []
GetForegroundWindow.restype = wintypes.HWND

SetForegroundWindow = windll.user32.SetForegroundWindow
SetForegroundWindow.argtypes = [
    wintypes.HWND,  # hWnd
]
SetForegroundWindow.restype = wintypes.BOOL



from os.path import commonprefix

from plover.formatting import (
    Case,
    META_ATTACH_FLAG,
    META_CARRY_CAPITALIZATION,
    has_word_boundary,
    rightmost_word,
)
from plover.orthography import add_suffix


def meta_attach(ctx, meta):
    action = ctx.new_action()
    begin = meta.startswith(META_ATTACH_FLAG)
    end = meta.endswith(META_ATTACH_FLAG)
    if not (begin or end):
        # If not specified, attach at both ends.
        meta = META_ATTACH_FLAG + meta + META_ATTACH_FLAG
        begin = end = True
    if begin:
        meta = meta[len(META_ATTACH_FLAG) :]
        action.prev_attach = True
    if end:
        meta = meta[: -len(META_ATTACH_FLAG)]
        action.next_attach = True
        action.word_is_finished = False
    last_word = ctx.last_action.word or ""
    if not meta:
        # We use an empty connection to indicate a "break" in the
        # application of orthography rules. This allows the
        # stenographer to tell Plover not to auto-correct a word.
        action.orthography = False
    elif (
        last_word
        and not meta.isspace()
        and ctx.last_action.orthography
        and begin
        and (not end or has_word_boundary(meta))
    ):
        new_word = add_suffix(last_word, meta)
        common_len = len(commonprefix([last_word, new_word]))
        replaced = last_word[common_len:]
        action.prev_replace = ctx.last_text(len(replaced))
        assert replaced.lower() == action.prev_replace.lower()
        last_word = last_word[:common_len]
        meta = new_word[common_len:]
    action.text = meta
    if action.prev_attach:
        action.word = rightmost_word(last_word + meta)
    return action


def meta_carry_capitalize(ctx, meta):
    # Meta format: ^~|content^ (attach flags are optional)
    action = ctx.new_action()
    if ctx.last_action.next_case == Case.CAP_FIRST_WORD:
        action.next_case = Case.CAP_FIRST_WORD
    begin = meta.startswith(META_ATTACH_FLAG)
    if begin:
        meta = meta[len(META_ATTACH_FLAG) :]
        action.prev_attach = True
    meta = meta[len(META_CARRY_CAPITALIZATION) :]
    end = meta.endswith(META_ATTACH_FLAG)
    if end:
        meta = meta[: -len(META_ATTACH_FLAG)]
        action.next_attach = True
        action.word_is_finished = False
    if meta or begin or end:
        action.text = meta
    return action



from collections import defaultdict

from packaging.utils import canonicalize_name

from plover.plugins_manager.package_index import find_plover_plugins_releases
from plover.plugins_manager.plugin_metadata import PluginMetadata


def list_plugins():
    plugins = defaultdict(list)
    for release in find_plover_plugins_releases():
        release_info = release["info"]
        plugin_metadata = PluginMetadata.from_dict(release_info)
        plugins[canonicalize_name(plugin_metadata.name)].append(plugin_metadata)
    plugins = {name: list(sorted(versions)) for name, versions in plugins.items()}
    return plugins



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

# TODO: maybe move this code into the StenoDictionary itself. The current saver
# structure is odd and awkward.
# TODO: write tests for this file

"""Common elements to all dictionary formats."""

from os.path import splitext
import functools
import threading

from plover.registry import registry


def _get_dictionary_class(filename):
    extension = splitext(filename)[1].lower()[1:]
    try:
        dict_module = registry.get_plugin("dictionary", extension).obj
    except KeyError:
        raise ValueError(
            "Unsupported extension: %s. Supported extensions: %s"
            % (
                extension,
                ", ".join(
                    plugin.name for plugin in registry.list_plugins("dictionary")
                ),
            )
        )
    return dict_module


def _locked(fn):
    lock = threading.Lock()

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        with lock:
            fn(*args, **kwargs)

    return wrapper


def _threaded(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        t = threading.Thread(target=fn, args=args, kwargs=kwargs)
        t.start()

    return wrapper


def create_dictionary(resource, threaded_save=True):
    """Create a new dictionary.

    The format is inferred from the extension.

    Note: the file is not created! The resulting dictionary save
    method must be called to finalize the creation on disk.
    """
    d = _get_dictionary_class(resource).create(resource)
    if threaded_save:
        d.save = _threaded(_locked(d.save))
    return d


def load_dictionary(resource, threaded_save=True):
    """Load a dictionary from a file.

    The format is inferred from the extension.
    """
    d = _get_dictionary_class(resource).load(resource)
    if not d.readonly and threaded_save:
        d.save = _threaded(_locked(d.save))
    return d



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Stenography translation.

This module handles translating streams of strokes in translations. Two classes
compose this module:

Translation -- A data model class that encapsulates a sequence of Stroke objects
in the context of a particular dictionary. The dictionary in question maps
stroke sequences to strings, which are typically words or phrases, but could
also be meta commands.

Translator -- A state machine that takes in a single Stroke object at a time and
emits one or more Translation objects based on a greedy conversion algorithm.

"""

from collections import namedtuple
import re

from plover.steno import Stroke
from plover.steno_dictionary import StenoDictionaryCollection
from plover.registry import registry
from plover import system


_ESCAPE_RX = re.compile("(\\\\[nrt]|[\n\r\t])")
_ESCAPE_REPLACEMENTS = {
    "\n": r"\n",
    "\r": r"\r",
    "\t": r"\t",
    r"\n": r"\\n",
    r"\r": r"\\r",
    r"\t": r"\\t",
}


def escape_translation(translation):
    return _ESCAPE_RX.sub(lambda m: _ESCAPE_REPLACEMENTS[m.group(0)], translation)


_UNESCAPE_RX = re.compile(r"((?<!\\)|\\)\\([nrt])")
_UNESCAPE_REPLACEMENTS = {
    r"\\n": r"\n",
    r"\\r": r"\r",
    r"\\t": r"\t",
    r"\n": "\n",
    r"\r": "\r",
    r"\t": "\t",
}


def unescape_translation(translation):
    return _UNESCAPE_RX.sub(lambda m: _UNESCAPE_REPLACEMENTS[m.group(0)], translation)


_LEGACY_MACROS_ALIASES = {
    "{*}": "retro_toggle_asterisk",
    "{*!}": "retro_delete_space",
    "{*?}": "retro_insert_space",
    "{*+}": "repeat_last_stroke",
}

_MACRO_RX = re.compile(r"=\w+(:|$)")

Macro = namedtuple("Macro", "name stroke cmdline")


def _mapping_to_macro(mapping, stroke):
    """Return a macro/stroke if mapping is one, or None otherwise."""
    macro, cmdline = None, ""
    if mapping is None:
        if stroke.is_correction:
            macro = "undo"
    else:
        if mapping in _LEGACY_MACROS_ALIASES:
            macro = _LEGACY_MACROS_ALIASES[mapping]
        elif _MACRO_RX.match(mapping):
            args = mapping[1:].split(":", 1)
            macro = args[0]
            if len(args) == 2:
                cmdline = args[1]
    return Macro(macro, stroke, cmdline) if macro else None


class Translation:
    """A data model for the mapping between a sequence of Strokes and a string.

    This class represents the mapping between a sequence of Stroke objects and
    a text string, typically a word or phrase. This class is used as the output
    from translation and the input to formatting. The class contains the
    following attributes:

    strokes -- A sequence of Stroke objects from which the translation is
    derived.

    rtfcre -- A tuple of RTFCRE strings representing the stroke list. This is
    used as the key in the translation mapping.

    english -- The value of the dictionary mapping given the rtfcre
    key, or None if no mapping exists.

    replaced -- A list of translations that were replaced by this one. If this
    translation is undone then it is replaced by these.

    formatting -- Information stored on the translation by the formatter for
    sticky state (e.g. capitalize next stroke) and to hold undo info.

    """

    def __init__(self, outline, translation):
        """Create a translation by looking up strokes in a dictionary.

        Arguments:

        outline -- A list of Stroke objects.

        translation -- A translation for the outline or None.

        """
        self.strokes = outline
        self.rtfcre = tuple(s.rtfcre for s in outline)
        self.english = translation
        self.replaced = []
        self.formatting = []
        self.is_retrospective_command = False

    def __eq__(self, other):
        return self.rtfcre == other.rtfcre and self.english == other.english

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        if self.english is None:
            translation = "None"
        else:
            translation = escape_translation(self.english)
            translation = '"%s"' % translation.replace('"', r"\"")
        return "Translation(%s : %s)" % (self.rtfcre, translation)

    def __repr__(self):
        return str(self)

    def __len__(self):
        if self.strokes is not None:
            return len(self.strokes)
        return 0

    def has_undo(self):
        # If there is no formatting then we're not dealing with a formatter
        # so all translations can be undone.
        # TODO: combos are not undoable but in some contexts they appear
        # as text. Should we provide a way to undo those? or is backspace
        # enough?
        if not self.formatting:
            return True
        if self.replaced:
            return True
        for a in self.formatting:
            if a.text or a.prev_replace:
                return True
        return False


class Translator:
    """Converts a stenotype key stream to a translation stream.

    An instance of this class serves as a state machine for processing key
    presses as they come off a stenotype machine. Key presses arrive in batches,
    each batch representing a single stenotype chord. The Translator class
    receives each chord as a Stroke and adds the Stroke to an internal,
    length-limited FIFO, which is then translated into a sequence of Translation
    objects. The resulting sequence of Translations is compared to those
    previously emitted by the state machine and a sequence of new Translations
    (some corrections and some new) is emitted.

    The internal Stroke FIFO is translated in a greedy fashion; the Translator
    finds a translation for the longest sequence of Strokes that starts with the
    oldest Stroke in the FIFO before moving on to newer Strokes that haven't yet
    been translated. In practical terms, this means that corrections are needed
    for cases in which a Translation comprises two or more Strokes, at least the
    first of which is a valid Translation in and of itself.

    For example, consider the case in which the first Stroke can be translated
    as 'cat'. In this case, a Translation object representing 'cat' will be
    emitted as soon as the Stroke is processed by the Translator. If the next
    Stroke is such that combined with the first they form 'catalogue', then the
    Translator will first issue a correction for the initial 'cat' Translation
    and then issue a new Translation for 'catalogue'.

    A Translator takes input via the translate method and provides translation
    output to every function that has registered via the add_callback method.

    """

    def __init__(self):
        self._undo_length = 0
        self._dictionary = None
        self.set_dictionary(StenoDictionaryCollection())
        self._listeners = set()
        self._state = _State()
        self._to_undo = []
        self._to_do = 0

    def translate(self, stroke):
        """Process a single stroke."""
        self.translate_stroke(stroke)
        self.flush()

    def set_dictionary(self, d):
        """Set the dictionary."""
        self._dictionary = d

    def get_dictionary(self):
        return self._dictionary

    def add_listener(self, callback):
        """Add a listener for translation outputs.

        Arguments:

        callback -- A function that takes: a list of translations to undo, a
        list of new translations to render, and a translation that is the
        context for the new translations.

        """
        self._listeners.add(callback)

    def remove_listener(self, callback):
        """Remove a listener added by add_listener."""
        self._listeners.remove(callback)

    def set_min_undo_length(self, n):
        """Set the minimum number of strokes that can be undone.

        The actual number may be larger depending on the translations in the
        dictionary.

        """
        self._undo_length = n
        self._resize_translations()

    def flush(self, extra_translations=None):
        """Process translations scheduled for undoing/doing.

        Arguments:

        extra_translations --  Extra translations to add to the list
                               of translation to do. Note: those will
                               not be saved to the state history.
        """
        if self._to_do:
            prev = self._state.prev(self._to_do)
            do = self._state.translations[-self._to_do :]
        else:
            prev = self._state.prev()
            do = []
        if extra_translations is not None:
            do.extend(extra_translations)
        undo = self._to_undo
        self._to_undo = []
        self._to_do = 0
        if undo or do:
            self._output(undo, do, prev)
        self._resize_translations()

    def _output(self, undo, do, prev):
        for callback in self._listeners:
            callback(undo, do, prev)

    def _resize_translations(self):
        self._state.restrict_size(max(self._dictionary.longest_key, self._undo_length))

    def get_state(self):
        """Get the state of the translator."""
        return self._state

    def set_state(self, state):
        """Set the state of the translator."""
        self._state = state

    def clear_state(self):
        """Reset the state of the translator."""
        self._state = _State()

    def translate_stroke(self, stroke):
        """Process a stroke.

        See the class documentation for details of how Stroke objects
        are converted to Translation objects.

        Arguments:

        stroke -- The Stroke object to process.

        """
        max_len = self._dictionary.longest_key
        mapping = self._lookup_with_prefix(max_len, self._state.translations, [stroke])
        macro = _mapping_to_macro(mapping, stroke)
        if macro is not None:
            self.translate_macro(macro)
            return
        t = (
            # No prefix lookups (note we avoid looking up [stroke] again).
            self._find_longest_match(2, max_len, stroke)
            or
            # Return [stroke] result if mapped.
            (mapping is not None and Translation([stroke], mapping))
            or
            # No direct match, try with suffixes.
            self._find_longest_match(1, max_len, stroke, system.SUFFIX_KEYS)
            or
            # Fallback to untranslate.
            Translation([stroke], None)
        )
        self.translate_translation(t)

    def translate_macro(self, macro):
        macro_fn = registry.get_plugin("macro", macro.name).obj
        macro_fn(self, macro.stroke, macro.cmdline)

    def translate_translation(self, t):
        self._undo(*t.replaced)
        self._do(t)

    def untranslate_translation(self, t):
        self._undo(t)
        self._do(*t.replaced)

    def _undo(self, *translations):
        for t in reversed(translations):
            assert t == self._state.translations.pop()
            if self._to_do:
                self._to_do -= 1
            else:
                self._to_undo.insert(0, t)

    def _do(self, *translations):
        self._state.translations.extend(translations)
        self._to_do += len(translations)

    def _find_longest_match(self, min_len, max_len, stroke, suffixes=()):
        """Find mapping with the longest series of strokes.

        min_len  -- Minimum number of strokes involved.
        max_len  -- Maximum number of strokes involved.
        stroke   -- The latest stroke.
        suffixes -- List of suffix keys to try.

        Return the corresponding translation, or None if no match is found.

        Note: the code either look for a direct match (empty suffix
        list), or assume the last stroke contains an implicit suffix
        and look for a corresponding match, but not both.
        """
        if suffixes:
            # Implicit suffix lookup, determine possible suffixes.
            suffixes = self._lookup_involved_suffixes(stroke, suffixes)
            if not suffixes:
                # No suffix involved, abort.
                return None
        # Figure out how much of the translation buffer can be involved in this
        # stroke and build the stroke list for translation.
        num_strokes = 1
        translation_count = 0
        for t in reversed(self._state.translations):
            num_strokes += len(t)
            if num_strokes > max_len:
                break
            translation_count += 1
        translation_index = len(self._state.translations) - translation_count
        translations = self._state.translations[translation_index:]
        # The new stroke can either create a new translation or replace
        # existing translations by matching a longer entry in the
        # dictionary.
        for i in range(len(translations) + 1):
            replaced = translations[i:]
            strokes = [s for t in replaced for s in t.strokes]
            strokes.append(stroke)
            if len(strokes) < min_len:
                continue
            mapping = self._lookup_with_prefix(
                max_len, translations[:i], strokes, suffixes
            )
            if mapping is not None:
                t = Translation(strokes, mapping)
                t.replaced = replaced
                return t
        return None

    def _lookup_strokes(self, strokes):
        """Look for a matching translation.

        strokes -- a list of Stroke instances.

        Return the resulting mapping.
        """
        return self._dictionary.lookup(tuple(s.rtfcre for s in strokes))

    def _lookup_with_suffix(self, strokes, suffixes=()):
        """Look for a matching translation.

        suffixes -- A list of (suffix stroke, suffix mapping) pairs to try.

        If the suffix list is empty, look for a direct match.

        Otherwise, assume the last stroke contains an implicit suffix,
        and look for a corresponding match.
        """
        if not suffixes:
            # No suffix, do a regular lookup.
            return self._lookup_strokes(strokes)
        for suffix_stroke, suffix_mapping in suffixes:
            assert suffix_stroke in strokes[-1]
            main_mapping = self._lookup_strokes(
                strokes[:-1] + [strokes[-1] - suffix_stroke]
            )
            if main_mapping is not None:
                return main_mapping + " " + suffix_mapping
        return None

    def _lookup_involved_suffixes(self, stroke, suffixes):
        """Find possible implicit suffixes for a stroke.

        stroke   -- The stroke to check for implicit suffixes.
        suffixes -- List of supported suffix keys.

        Return a list of (suffix_stroke, suffix_mapping) pairs.
        """
        possible_suffixes = []
        for suffix_stroke in map(Stroke, suffixes):
            if suffix_stroke not in stroke:
                continue
            suffix_mapping = self._lookup_strokes((suffix_stroke,))
            if suffix_mapping is None:
                continue
            possible_suffixes.append((suffix_stroke, suffix_mapping))
        return possible_suffixes

    def lookup(self, strokes, suffixes=()):
        result = self._lookup_strokes(strokes)
        if result is not None:
            return result
        suffixes = self._lookup_involved_suffixes(strokes[-1], suffixes)
        if not suffixes:
            return None
        return self._lookup_with_suffix(strokes, suffixes)

    @staticmethod
    def _previous_word_is_finished(last_translations):
        if not last_translations:
            return True
        formatting = last_translations[-1].formatting
        if not formatting:
            return True
        return formatting[-1].word_is_finished

    def _lookup_with_prefix(self, max_len, last_translations, strokes, suffixes=()):
        if len(strokes) < max_len and self._previous_word_is_finished(
            last_translations
        ):
            mapping = self._lookup_with_suffix(
                [Stroke.PREFIX_STROKE] + strokes, suffixes
            )
            if mapping is not None:
                return mapping
        if len(strokes) <= max_len:
            return self._lookup_with_suffix(strokes, suffixes)
        return None


class _State:
    """An object representing the current state of the translator state machine.

    Attributes:

    translations -- A list of all previous translations that are still undoable.

    tail -- The oldest translation still saved but is no longer undoable.

    """

    def __init__(self):
        self.translations = []
        self.tail = None

    def prev(self, count=None):
        """Get the most recent translations."""
        if count is not None:
            prev = self.translations[:-count]
        else:
            prev = self.translations
        if prev:
            return prev
        if self.tail is not None:
            return [self.tail]
        return None

    def restrict_size(self, n):
        """Reduce the history of translations to n."""
        stroke_count = 0
        translation_count = 0
        for t in reversed(self.translations):
            stroke_count += len(t)
            translation_count += 1
            if stroke_count >= n:
                break
        translation_index = len(self.translations) - translation_count
        if translation_index:
            self.tail = self.translations[translation_index - 1]
        del self.translations[:translation_index]



from threading import Event

from plover.oslayer.keyboardcontrol import KeyboardEmulation

from plover.gui_none.engine import Engine


def show_error(title, message):
    print("%s: %s" % (title, message))


def main(config, controller):
    engine = Engine(config, controller, KeyboardEmulation())
    if not engine.load_config():
        return 3
    quitting = Event()
    engine.hook_connect("quit", quitting.set)
    engine.start()
    try:
        quitting.wait()
    except KeyboardInterrupt:
        engine.quit()
    return engine.join()



#!/usr/bin/env python3

import glob
import shutil
import sys
import os


def trim(directory, patterns_file, verbose=True, dry_run=False):
    if dry_run:
        verbose = True
    # Build list of patterns.
    pattern_list = []
    exclude_list = []
    subdir = directory
    with open(patterns_file) as fp:
        for line in fp:
            line = line.strip()
            # Empty line, ignore.
            if not line:
                continue
            # Comment, ignore.
            if line.startswith("#"):
                continue
            # Sub-directory change.
            if line.startswith(":"):
                subdir = os.path.join(directory, line[1:])
                continue
            # Pattern (relative to current sub-directory).
            if line.startswith("!"):
                exclude_list.append(os.path.join(subdir, line[1:]))
            else:
                pattern_list.append(os.path.join(subdir, line))
    # Create list of files to keep based on exclusion list.
    to_keep = set()
    for pattern in exclude_list:
        to_keep.update(glob.glob(pattern, recursive=True))
    # Trim directory tree.
    for pattern in pattern_list:
        for path in reversed(glob.glob(pattern, recursive=True)):
            if path in to_keep:
                continue
            if verbose:
                print("removing", path)
            if dry_run:
                continue
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)


if __name__ == "__main__":
    directory, patterns_file = sys.argv[1:]
    trim(directory, patterns_file)



from pathlib import Path


def patch_ports_info(port_list):
    """Patch serial ports info to use device-by-id links."""
    try:
        device_by_id = {
            str(device.resolve()): str(device)
            for device in Path("/dev/serial/by-id").iterdir()
        }
    except FileNotFoundError:
        device_by_id = {}
    for port_info in port_list:
        port_info.device = device_by_id.get(port_info.device, port_info.device)
    return port_list



from plover.translation import Translation
from plover.oslayer.config import PLATFORM


if PLATFORM == "mac":
    BACK_STRING = "{#Alt_L(BackSpace)}{^}"
else:
    BACK_STRING = "{#Control_L(BackSpace)}{^}"


def undo(translator, stroke, cmdline):
    assert not cmdline
    for t in reversed(translator.get_state().translations):
        translator.untranslate_translation(t)
        if t.has_undo():
            return
    # There is no more buffer to delete from -- remove undo and add a
    # stroke that removes last word on the user's OS, but don't add it
    # to the state history.
    translator.flush([Translation([stroke], BACK_STRING)])



from threading import Thread
import atexit
import html
import os
import sys

from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtWidgets import (
    QDialog,
    QMessageBox,
    QTableWidgetItem,
    QInputDialog,
    QWidget,
)

from plover.gui_qt.tool import Tool
from plover.gui_qt.info_browser import InfoBrowser
from plover.gui_qt.plugins_manager_ui import Ui_PluginsManager
from plover.gui_qt.run_dialog import RunDialog
from plover.plugins_manager.registry import Registry
from plover.plugins_manager.utils import description_to_html
from plover.plugins_manager.__main__ import pip


class PluginsManager(Tool, Ui_PluginsManager):
    TITLE = "Plugins Manager"
    ROLE = "plugins_manager"
    ICON = ":/resources/plugins_manager.svg"

    # We use a class instance so the state is persistent
    # accross different executions of the dialog when
    # the user does not restart.
    _packages = None
    _packages_updated = Signal()

    def __init__(self, engine):
        super().__init__(engine)
        self.setupUi(self)
        self.uninstall_button.setEnabled(False)
        self.install_button.setEnabled(False)
        self._engine = engine
        self.info = InfoBrowser()
        self.info_frame.layout().addWidget(self.info)
        self.table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
        self._packages_updated.connect(self._on_packages_updated)
        if self._packages is None:
            PluginsManager._packages = Registry()
        self._on_packages_updated()
        self.refresh()

    def _need_restart(self):
        for state in self._packages:
            if state.status in ("removed", "updated"):
                return True
        return False

    def _on_packages_updated(self):
        self.restart_button.setEnabled(self._need_restart())
        self.progress.hide()
        self.refresh_button.show()
        self._update_table()
        self.setEnabled(True)

    def _update_table(self):
        self.table.setCurrentItem(None)
        self.table.setSortingEnabled(False)
        self.table.setRowCount(len(self._packages))
        for row, state in enumerate(self._packages):
            for column, attr in enumerate("status name version summary".split()):
                item = QTableWidgetItem(getattr(state, attr, "N/A"))
                item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
                self.table.setItem(row, column, item)
        self.table.resizeColumnsToContents()
        self.table.setSortingEnabled(True)

    def _get_state(self, row):
        name = self.table.item(row, 1).data(Qt.ItemDataRole.DisplayRole)
        return self._packages[name]

    def _get_selection(self):
        can_install = []
        can_uninstall = []
        for item in self.table.selectedItems():
            if item.column() != 0:
                continue
            state = self._get_state(item.row())
            if state.status in ("installed", "updated"):
                can_uninstall.append(state.name)
            elif state.status in ("outdated",):
                can_uninstall.append(state.name)
                can_install.append(state.name)
            elif state.status != "unsupported" and state.latest:
                can_install.append(state.name)
        return can_install, can_uninstall

    @staticmethod
    def _run(args):
        dialog = RunDialog(args, popen=pip)
        code = dialog.exec()
        # dialog.destroy()
        return code

    @Slot()
    def handle_selection_change(self):
        can_install, can_uninstall = self._get_selection()
        self.uninstall_button.setEnabled(bool(can_uninstall))
        self.install_button.setEnabled(bool(can_install))
        self._clear_info()
        current_item = self.table.currentItem()
        if current_item is None:
            return
        metadata = self._get_state(current_item.row()).metadata
        if metadata is None:
            return
        prologue = "<h1>%s (%s)</h1>" % (
            html.escape(metadata.name),
            html.escape(metadata.version),
        )
        if metadata.author and metadata.author_email:
            prologue += '<p><b>Author: </b><a href="mailto:%s">%s</a></p>' % (
                html.escape(metadata.author_email),
                html.escape(metadata.author),
            )
        if metadata.home_page:
            prologue += '<p><b>Home page: </b><a href="%s">%s</a></p>' % (
                metadata.home_page,
                html.escape(metadata.home_page),
            )
        prologue += "<hr>"
        if metadata.description:
            description = metadata.description
            description_content_type = metadata.description_content_type
        else:
            description = metadata.summary
            description_content_type = None
        css, description = description_to_html(description, description_content_type)
        self.info.setHtml(css + prologue + description)

    @Slot()
    def restart(self):
        if self._engine is not None:
            self._engine.restart()
        else:
            atexit._run_exitfuncs()
            args = [sys.executable, "-m", __spec__.name]
            os.execv(args[0], args)

    def _update_packages(self):
        self._packages.update()
        self._packages_updated.emit()

    def _clear_info(self):
        self.info.setHtml("")

    @Slot()
    def refresh(self):
        Thread(target=self._update_packages).start()
        self._clear_info()
        self.setEnabled(False)
        self.refresh_button.hide()
        self.progress.show()

    @Slot()
    def install_from_git(self):
        url, ok = QInputDialog.getText(
            self,
            "Install from Git repo",
            "<b>WARNING: Installing plugins is a security risk.<br>"
            "A plugin from a Git repo can contain malicious code.<br>"
            "Only install it if you got it from a trusted source.</b><br><br>"
            "Enter repository link for plugin<br>"
            "(will look similar to "
            "https://github.com/user/repository.git): <br>",
        )
        if not ok or not url:
            return

        code = self._run(["install"] + ["git+" + url])
        if code == QDialog.DialogCode.Accepted:
            self._update_table()
            self.restart_button.setEnabled(True)

    @Slot()
    def install_selected_package(self):
        packages = self._get_selection()[0]
        if (
            QMessageBox.warning(
                self,
                "Install " + ", ".join(packages),
                "Installing plugins is a <b>security risk</b>. "
                "A plugin can contain virus/malware. "
                "Only install it if you got it from a trusted source."
                " Are you sure you want to proceed?",
                buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                defaultButton=QMessageBox.StandardButton.No,
            )
            != QMessageBox.StandardButton.Yes
        ):
            return
        code = self._run(
            ["install"] + [self._packages[name].latest.requirement for name in packages]
        )
        if code == QDialog.DialogCode.Accepted:
            for name in packages:
                state = self._packages[name]
                state.current = state.latest
            self._update_table()
            self.restart_button.setEnabled(True)

    @Slot()
    def uninstall_selected_package(self):
        packages = self._get_selection()[1]
        code = self._run(["uninstall", "-y"] + packages)
        if code == QDialog.DialogCode.Accepted:
            for name in packages:
                state = self._packages[name]
                state.current = None
            self._update_table()
            self.restart_button.setEnabled(True)


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])
    dlg = PluginsManager(None)
    dlg.show()
    app.exec()



from distutils import sysconfig
import sys

print(
    "; ".join(
        "py%s=%r" % (k, v)
        for k, v in sorted(
            {
                "exe": sys.executable,
                "prefix": getattr(sys, "base_prefix", sys.prefix),
                "version": sysconfig.get_python_version(),
                "include": sysconfig.get_python_inc(prefix=""),
                "ldlib": sysconfig.get_config_var("LDLIBRARY"),
                "py3lib": sysconfig.get_config_var("PY3LIBRARY"),
                "stdlib": sysconfig.get_python_lib(standard_lib=1, prefix=""),
                "purelib": sysconfig.get_python_lib(plat_specific=0, prefix=""),
                "platlib": sysconfig.get_python_lib(plat_specific=1, prefix=""),
            }.items()
        )
    )
)



# coding: utf-8

import threading
from time import sleep
from queue import Queue

from Quartz import (
    CFMachPortCreateRunLoopSource,
    CFMachPortInvalidate,
    CFRunLoopAddSource,
    CFRunLoopSourceInvalidate,
    CFRunLoopGetCurrent,
    CFRunLoopRun,
    CFRunLoopStop,
    CFRelease,
    CGEventCreateKeyboardEvent,
    CGEventGetFlags,
    CGEventGetIntegerValueField,
    CGEventKeyboardSetUnicodeString,
    CGEventMaskBit,
    CGEventPost,
    CGEventSetFlags,
    CGEventSourceCreate,
    CGEventTapCreate,
    CGEventTapEnable,
    kCFRunLoopCommonModes,
    kCGEventFlagMaskAlternate,
    kCGEventFlagMaskCommand,
    kCGEventFlagMaskControl,
    kCGEventFlagMaskNonCoalesced,
    kCGEventFlagMaskNumericPad,
    kCGEventFlagMaskSecondaryFn,
    kCGEventFlagMaskShift,
    kCGEventKeyDown,
    kCGEventKeyUp,
    kCGEventTapDisabledByTimeout,
    kCGEventTapOptionDefault,
    kCGHeadInsertEventTap,
    kCGKeyboardEventKeycode,
    kCGSessionEventTap,
    kCGEventSourceStateHIDSystemState,
    NSEvent,
    NSSystemDefined,
)

from plover import log
from plover.key_combo import add_modifiers_aliases, parse_key_combo, KEYNAME_TO_CHAR
from plover.machine.keyboard_capture import Capture
from plover.output.keyboard import GenericKeyboardEmulation

from .keyboardlayout import KeyboardLayout


BACK_SPACE = 51

NX_KEY_OFFSET = 65536

NX_KEYS = {
    "AudioRaiseVolume": 0,
    "AudioLowerVolume": 1,
    "MonBrightnessUp": 2,
    "MonBrightnessDown": 3,
    "AudioMute": 7,
    "Num_Lock": 10,
    "Eject": 14,
    "AudioPause": 16,
    "AudioPlay": 16,
    "AudioNext": 17,
    "AudioPrev": 18,
    "AudioRewind": 20,
    "KbdBrightnessUp": 21,
    "KbdBrightnessDown": 22,
}

KEYNAME_TO_KEYCODE = {
    "fn": 63,
    "alt_l": 58,
    "alt_r": 61,
    "control_l": 59,
    "control_r": 62,
    "shift_l": 56,
    "shift_r": 60,
    "super_l": 55,
    "super_r": 55,
    "caps_lock": 57,
    "return": 36,
    "tab": 48,
    "backspace": BACK_SPACE,
    "delete": 117,
    "escape": 53,
    "clear": 71,
    "up": 126,
    "down": 125,
    "left": 123,
    "right": 124,
    "page_up": 116,
    "page_down": 121,
    "home": 115,
    "end": 119,
    "f1": 122,
    "f2": 120,
    "f3": 99,
    "f4": 118,
    "f5": 96,
    "f6": 97,
    "f7": 98,
    "f8": 100,
    "f9": 101,
    "f10": 109,
    "f11": 103,
    "f12": 111,
    "f13": 105,
    "f14": 107,
    "f15": 113,
    "f16": 106,
    "f17": 64,
    "f18": 79,
    "f19": 80,
    "f20": 90,
    "kp_0": 82,
    "kp_1": 83,
    "kp_2": 84,
    "kp_3": 85,
    "kp_4": 86,
    "kp_5": 87,
    "kp_6": 88,
    "kp_7": 89,
    "kp_8": 91,
    "kp_9": 92,
    "kp_add": 69,
    "kp_decimal": 65,
    "kp_delete": 71,
    "kp_divide": 75,
    "kp_enter": 76,
    "kp_equal": 81,
    "kp_multiply": 67,
    "kp_subtract": 78,
}
for name, code in NX_KEYS.items():
    KEYNAME_TO_KEYCODE[name.lower()] = code + NX_KEY_OFFSET
add_modifiers_aliases(KEYNAME_TO_KEYCODE)

DEADKEY_SYMBOLS = {
    "dead_acute": "´",
    "dead_circumflex": "^",
    "dead_diaeresis": "¨",
    "dead_grave": "`",
    "dead_tilde": "~",
}


def down(seq):
    return [(x, True) for x in seq]


def up(seq):
    return [(x, False) for x in reversed(seq)]


def down_up(seq):
    return down(seq) + up(seq)


# Maps from keycodes to corresponding event masks.
MODIFIER_KEYS_TO_MASKS = {
    58: kCGEventFlagMaskAlternate,
    61: kCGEventFlagMaskAlternate,
    # As of Sierra we *require* secondary fn to get control to work properly.
    59: kCGEventFlagMaskControl,
    62: kCGEventFlagMaskControl,
    56: kCGEventFlagMaskShift,
    60: kCGEventFlagMaskShift,
    55: kCGEventFlagMaskCommand,
    63: kCGEventFlagMaskSecondaryFn,
}

OUTPUT_SOURCE = CGEventSourceCreate(kCGEventSourceStateHIDSystemState)

# For the purposes of this class, we're only watching these keys.
# We could calculate the keys, but our default layout would be misleading:
#
#     KEYCODE_TO_KEY = {keycode: mac_keycode.CharForKeycode(keycode) \
#             for keycode in range(127)}
KEYCODE_TO_KEY = {
    122: "F1",
    120: "F2",
    99: "F3",
    118: "F4",
    96: "F5",
    97: "F6",
    98: "F7",
    100: "F8",
    101: "F9",
    109: "F10",
    103: "F11",
    111: "F12",
    50: "`",
    29: "0",
    18: "1",
    19: "2",
    20: "3",
    21: "4",
    23: "5",
    22: "6",
    26: "7",
    28: "8",
    25: "9",
    27: "-",
    24: "=",
    12: "q",
    13: "w",
    14: "e",
    15: "r",
    17: "t",
    16: "y",
    32: "u",
    34: "i",
    31: "o",
    35: "p",
    33: "[",
    30: "]",
    42: "\\",
    0: "a",
    1: "s",
    2: "d",
    3: "f",
    5: "g",
    4: "h",
    38: "j",
    40: "k",
    37: "l",
    41: ";",
    39: "'",
    6: "z",
    7: "x",
    8: "c",
    9: "v",
    11: "b",
    45: "n",
    46: "m",
    43: ",",
    47: ".",
    44: "/",
    49: "space",
    BACK_SPACE: "BackSpace",
    117: "Delete",
    125: "Down",
    119: "End",
    53: "Escape",
    115: "Home",
    123: "Left",
    121: "Page_Down",
    116: "Page_Up",
    36: "Return",
    124: "Right",
    48: "Tab",
    126: "Up",
}


def keycode_needs_fn_mask(keycode):
    return (
        keycode >= KEYNAME_TO_KEYCODE["escape"]
        and keycode not in MODIFIER_KEYS_TO_MASKS
    )


class KeyboardCaptureLoop:
    def __init__(self, callback):
        self._callback = callback
        self._loop = None
        self._source = None
        self._tap = None
        self._thread = None

    def _run(self, loop_is_set):
        self._loop = CFRunLoopGetCurrent()
        loop_is_set.set()
        del loop_is_set
        CFRunLoopAddSource(self._loop, self._source, kCFRunLoopCommonModes)
        CFRunLoopRun()

    def start(self):
        self._tap = CGEventTapCreate(
            kCGSessionEventTap,
            kCGHeadInsertEventTap,
            kCGEventTapOptionDefault,
            CGEventMaskBit(kCGEventKeyDown) | CGEventMaskBit(kCGEventKeyUp),
            self._callback,
            None,
        )
        if self._tap is None:
            # TODO: See if there is a nice way to show
            # the user what's needed (or do it for them).
            raise Exception("Enable access for assistive devices.")
        self._source = CFMachPortCreateRunLoopSource(None, self._tap, 0)
        loop_is_set = threading.Event()
        self._thread = threading.Thread(
            target=self._run, args=(loop_is_set,), name="KeyboardCaptureLoop"
        )
        self._thread.start()
        loop_is_set.wait()
        self.toggle_tap(True)

    def toggle_tap(self, enabled):
        CGEventTapEnable(self._tap, enabled)

    def cancel(self):
        if self._loop is not None:
            CFRunLoopStop(self._loop)
        if self._thread is not None:
            self._thread.join()
        if self._tap is not None:
            CFMachPortInvalidate(self._tap)
            CFRelease(self._tap)
        if self._source is not None:
            CFRunLoopSourceInvalidate(self._source)


class KeyboardCapture(Capture):
    _KEYBOARD_EVENTS = {kCGEventKeyDown, kCGEventKeyUp}

    # Don't ignore Fn and Numeric flags so we can handle
    # the arrow and extended (home, end, etc...) keys.
    _PASSTHROUGH_MODIFIERS = ~(
        kCGEventFlagMaskNumericPad
        | kCGEventFlagMaskSecondaryFn
        | kCGEventFlagMaskNonCoalesced
    )

    def __init__(self):
        super().__init__()
        self.key_down = lambda key: None
        self.key_up = lambda key: None
        self._suppressed_keys = set()
        self._event_queue = Queue()
        self._event_thread = None
        self._capture_loop = None

    def _callback(self, proxy, event_type, event, reference):
        """
        Returning the event means that it is passed on
        for further processing by others.

        Returning None means that the event is intercepted.

        Delaying too long in returning appears to cause the
        system to ignore the tap forever after
        (https://github.com/openstenoproject/plover/issues/484#issuecomment-214743466).

        This motivates pushing callbacks to the other side
        of a queue of received events, so that we can return
        from this callback as soon as possible.
        """
        if event_type not in self._KEYBOARD_EVENTS:
            # Don't pass on meta events meant for this event tap.
            if event_type == kCGEventTapDisabledByTimeout:
                # Re-enable the tap and hope we act faster next time.
                self._capture_loop.toggle_tap(True)
                log.warning(
                    "Keystrokes may have been missed, the"
                    "keyboard event tap has been re-enabled."
                )
            return None
        if CGEventGetFlags(event) & self._PASSTHROUGH_MODIFIERS:
            # Don't intercept or suppress the event if it has
            # modifiers and the whole keyboard is not suppressed.
            return event
        keycode = CGEventGetIntegerValueField(event, kCGKeyboardEventKeycode)
        key = KEYCODE_TO_KEY.get(keycode)
        if key is None:
            # Not a supported key, don't intercept or suppress.
            return event
        # Notify supported key event.
        self._event_queue.put_nowait((key, event_type))
        # And suppressed it if asked too.
        return None if key in self._suppressed_keys else event

    def start(self):
        self._capture_loop = KeyboardCaptureLoop(self._callback)
        self._capture_loop.start()
        self._event_thread = threading.Thread(name="KeyboardCapture", target=self._run)
        self._event_thread.start()

    def cancel(self):
        if self._event_thread is not None:
            self._event_queue.put(None)
            self._event_thread.join()
        if self._capture_loop is not None:
            self._capture_loop.cancel()

    def suppress(self, suppressed_keys=()):
        self._suppressed_keys = set(suppressed_keys)

    def _run(self):
        while True:
            event = self._event_queue.get()
            if event is None:
                return
            key, event_type = event
            if event_type == kCGEventKeyUp:
                self.key_up(key)
            else:
                self.key_down(key)


class KeyboardEmulation(GenericKeyboardEmulation):
    RAW_PRESS, STRING_PRESS = range(2)

    def __init__(self):
        super().__init__()
        self._layout = KeyboardLayout()

    def send_backspaces(self, count):
        for _ in self.with_delay(range(count)):
            backspace_down = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, BACK_SPACE, True)
            backspace_up = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, BACK_SPACE, False)
            CGEventPost(kCGSessionEventTap, backspace_down)
            CGEventPost(kCGSessionEventTap, backspace_up)

    def send_string(self, string):
        # Key plan will store the type of output
        # (raw keycodes versus setting string)
        # and the list of keycodes or the goal character.
        key_plan = []

        # apply_raw's properties are used to store
        # the current keycode sequence,
        # and add the sequence to the key plan when called as a function.
        def apply_raw():
            if hasattr(apply_raw, "sequence") and len(apply_raw.sequence) != 0:
                apply_raw.sequence.extend(apply_raw.release_modifiers)
                key_plan.append((self.RAW_PRESS, apply_raw.sequence))
            apply_raw.sequence = []
            apply_raw.release_modifiers = []

        apply_raw()

        last_modifier = None
        for c in string:
            for keycode, modifier in self._layout.char_to_key_sequence(c):
                if keycode is not None:
                    if modifier is not last_modifier:
                        # Flush on modifier change.
                        apply_raw()
                        last_modifier = modifier
                        modifier_keycodes = self._modifier_to_keycodes(modifier)
                        apply_raw.sequence.extend(down(modifier_keycodes))
                        apply_raw.release_modifiers.extend(up(modifier_keycodes))
                    apply_raw.sequence.extend(down_up([keycode]))
                else:
                    # Flush on type change.
                    last_modifier = None
                    apply_raw()
                    key_plan.append((self.STRING_PRESS, c))
        # Flush after string is complete.
        apply_raw()

        # We have a key plan for the whole string, grouping modifiers.
        for press_type, sequence in self.with_delay(key_plan):
            if press_type is self.STRING_PRESS:
                self._send_string_press(sequence)
            elif press_type is self.RAW_PRESS:
                self._send_sequence(sequence)

    @staticmethod
    def _send_string_press(c):
        event = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, 0, True)
        KeyboardEmulation._set_event_string(event, c)
        CGEventPost(kCGSessionEventTap, event)
        event = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, 0, False)
        KeyboardEmulation._set_event_string(event, c)
        CGEventPost(kCGSessionEventTap, event)

    def send_key_combination(self, combo):
        def name_to_code(name):
            # Static key codes
            code = KEYNAME_TO_KEYCODE.get(name)
            if code is not None:
                pass
            # Dead keys
            elif name.startswith("dead_"):
                code, mod = self._layout.deadkey_symbol_to_key_sequence(
                    DEADKEY_SYMBOLS.get(name)
                )[0]
            # Normal keys
            else:
                char = KEYNAME_TO_CHAR.get(name, name)
                code, mods = self._layout.char_to_key_sequence(char)[0]
            return code

        # Parse and validate combo.
        key_events = parse_key_combo(combo, name_to_code)
        # Send events...
        self._send_sequence(key_events)

    @staticmethod
    def _modifier_to_keycodes(modifier):
        keycodes = []
        if modifier & 16:
            keycodes.append(KEYNAME_TO_KEYCODE["control_r"])
        if modifier & 8:
            keycodes.append(KEYNAME_TO_KEYCODE["alt_r"])
        if modifier & 2:
            keycodes.append(KEYNAME_TO_KEYCODE["shift_r"])
        if modifier & 1:
            keycodes.append(KEYNAME_TO_KEYCODE["super_r"])
        return keycodes

    @staticmethod
    def _set_event_string(event, s):
        nb_utf16_codepoints = len(s.encode("utf-16-le")) // 2
        CGEventKeyboardSetUnicodeString(event, nb_utf16_codepoints, s)

    MODS_MASK = (
        kCGEventFlagMaskAlternate
        | kCGEventFlagMaskControl
        | kCGEventFlagMaskShift
        | kCGEventFlagMaskCommand
        | kCGEventFlagMaskSecondaryFn
    )

    @staticmethod
    def _get_media_event(key_id, key_down):
        # Credit: https://gist.github.com/fredrikw/4078034
        flags = 0xA00 if key_down else 0xB00
        return NSEvent.otherEventWithType_location_modifierFlags_timestamp_windowNumber_context_subtype_data1_data2_(  # nopep8: We can't make this line shorter!
            NSSystemDefined, (0, 0), flags, 0, 0, 0, 8, (key_id << 16) | flags, -1
        ).CGEvent()

    def _send_sequence(self, sequence):
        # There is a bug in the event system that seems to cause inconsistent
        # modifiers on key events:
        # http://stackoverflow.com/questions/2008126/cgeventpost-possible-bug-when-simulating-keyboard-events
        # My solution is to manage the state myself.
        # I'm not sure how to deal with caps lock.
        # If mods_flags is not zero at the end then bad things might happen.
        mods_flags = 0

        for keycode, key_down in self.with_delay(sequence):
            if keycode >= NX_KEY_OFFSET:
                # Handle media (NX) key.
                event = KeyboardEmulation._get_media_event(
                    keycode - NX_KEY_OFFSET, key_down
                )
            else:
                # Handle regular keycode.
                if not key_down and keycode in MODIFIER_KEYS_TO_MASKS:
                    mods_flags &= ~MODIFIER_KEYS_TO_MASKS[keycode]

                if key_down and keycode_needs_fn_mask(keycode):
                    mods_flags |= kCGEventFlagMaskSecondaryFn

                event = CGEventCreateKeyboardEvent(OUTPUT_SOURCE, keycode, key_down)

                if key_down and keycode not in MODIFIER_KEYS_TO_MASKS:
                    event_flags = CGEventGetFlags(event)
                    # Add wanted flags, remove unwanted flags.
                    goal_flags = (
                        event_flags & ~KeyboardEmulation.MODS_MASK
                    ) | mods_flags
                    if event_flags != goal_flags:
                        CGEventSetFlags(event, goal_flags)

                    # Half millisecond pause after key down.
                    sleep(0.0005)

                if key_down and keycode in MODIFIER_KEYS_TO_MASKS:
                    mods_flags |= MODIFIER_KEYS_TO_MASKS[keycode]

                if not key_down and keycode_needs_fn_mask(keycode):
                    mods_flags &= ~kCGEventFlagMaskSecondaryFn

            CGEventPost(kCGSessionEventTap, event)



from PySide6.QtCore import (
    QThread,
    Signal,
)

from plover.engine import StenoEngine
from plover.oslayer.config import PLATFORM


class Engine(StenoEngine, QThread):
    # Signals.
    signal_stroked = Signal(object)
    signal_translated = Signal(object, object)
    signal_machine_state_changed = Signal(str, str)
    signal_output_changed = Signal(bool)
    signal_config_changed = Signal(object)
    signal_dictionary_state_changed = Signal(
        str, object
    )  # Some dictionary has finished loading. Refer to class DictionaryLoadingManager for argument description.
    signal_dictionaries_loaded = Signal(
        object
    )  # All dictionaries are loaded. Argument is a StenoDictionaryCollection instance.
    signal_send_string = Signal(str)
    signal_send_backspaces = Signal(int)
    signal_send_key_combination = Signal(str)
    signal_add_translation = Signal()
    signal_focus = Signal()
    signal_configure = Signal()
    signal_lookup = Signal()
    signal_suggestions = Signal()
    signal_quit = Signal()

    def __init__(self, config, controller, keyboard_emulation):
        StenoEngine.__init__(self, config, controller, keyboard_emulation)
        QThread.__init__(self)
        self._signals = {}
        for hook in self.HOOKS:
            signal = getattr(self, "signal_" + hook)
            self.hook_connect(hook, signal.emit)
            self._signals[hook] = signal

    def _in_engine_thread(self):
        return self.currentThread() == self

    def start(self):
        QThread.start(self)
        StenoEngine.start(self)

    def join(self):
        QThread.wait(self)
        return self.code

    def run(self):
        if PLATFORM == "mac":
            import appnope

            appnope.nope()
        super().run()

    def signal_connect(self, name, callback):
        self._signals[name].connect(callback)



import contextlib
import importlib
import os
import subprocess
import sys

from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
import setuptools
from importlib.metadata import distribution, PackageNotFoundError


class Command(setuptools.Command):
    def build_in_place(self):
        self.run_command("build_py")
        self.reinitialize_command("build_ext", inplace=1)
        self.run_command("build_ext")

    @contextlib.contextmanager
    def project_on_sys_path(self, build=True):
        ei_cmd = self.get_finalized_command("egg_info")
        if build:
            self.build_in_place()
        else:
            ei_cmd.run()
        old_path = sys.path[:]
        old_modules = sys.modules.copy()
        try:
            sys.path.insert(0, os.path.abspath(ei_cmd.egg_base))
            dist = distribution(ei_cmd.egg_name)
            if dist is None:
                raise PackageNotFoundError(f"Package {ei_cmd.egg_name} not found")
            yield
        finally:
            sys.path[:] = old_path
            sys.modules.clear()
            sys.modules.update(old_modules)

    def bdist_wheel(self):
        """Run bdist_wheel and return resulting wheel file path."""
        whl_cmd = self.get_finalized_command("bdist_wheel")
        whl_cmd.run()
        for cmd, py_version, dist_path in whl_cmd.distribution.dist_files:
            if cmd == "bdist_wheel":
                return dist_path
        raise Exception("could not find wheel path")


# i18n support. {{{


def babel_options(package, resource_dir=None):
    if resource_dir is None:
        localedir = "%s/messages" % package
    else:
        localedir = "%s/%s" % (package, resource_dir)
    template = "%s/%s.pot" % (localedir, package)
    return {
        "compile_catalog": {
            "domain": package,
            "directory": localedir,
        },
        "extract_messages": {
            "add_comments": ["i18n:"],
            "output_file": template,
            "strip_comments": True,
        },
        "init_catalog": {
            "domain": package,
            "input_file": template,
            "output_dir": localedir,
        },
        "update_catalog": {
            "domain": package,
            "output_dir": localedir,
        },
    }


# }}}

# UI generation. {{{


class BuildUi(Command):
    description = "build UI files"
    user_options = [
        ("force", "f", "force re-generation of all UI files"),
    ]

    hooks = """
    plover_build_utils.pyqt:no_autoconnection
    """.split()

    def initialize_options(self):
        self.force = False

    def finalize_options(self):
        pass

    def _build_ui(self, src):
        dst = os.path.splitext(src)[0] + "_ui.py"
        if (
            not self.force
            and os.path.exists(dst)
            and os.path.getmtime(dst) >= os.path.getmtime(src)
        ):
            return
        if self.verbose:
            print("generating", dst)

        subprocess.check_call(["pyside6-uic", "--from-imports", src, "-o", dst])

        for hook in self.hooks:
            mod_name, attr_name = hook.split(":")
            mod = importlib.import_module(mod_name)
            hook_fn = getattr(mod, attr_name)
            with open(dst, "r") as fp:
                contents = fp.read()
            contents = hook_fn(contents)
            with open(dst, "w") as fp:
                fp.write(contents)

    def _build_resources(self, src):
        dst = os.path.splitext(src)[0] + "_rc.py"
        # Skip if up‑to‑date unless --force was requested
        if (
            not self.force
            and os.path.exists(dst)
            and os.path.getmtime(dst) >= os.path.getmtime(src)
        ):
            return
        if self.verbose:
            print("compiling", src, "->", dst)
        subprocess.check_call(["pyside6-rcc", "-o", dst, src])

    def run(self):
        self.run_command("egg_info")
        std_hook_prefix = __package__ + ".pyqt:"
        hooks_info = [
            h[len(std_hook_prefix) :] if h.startswith(std_hook_prefix) else h
            for h in self.hooks
        ]
        if self.verbose:
            print("generating UI using hooks:", ", ".join(hooks_info))
        ei_cmd = self.get_finalized_command("egg_info")

        for src in ei_cmd.filelist.files:
            if src.endswith(".qrc"):
                self._build_resources(src)
            if src.endswith(".ui"):
                self._build_ui(src)


# }}}


# Patched `build_py` command. {{{


class BuildPy(build_py):
    build_dependencies = []

    def run(self):
        for command in self.build_dependencies:
            self.run_command(command)
        build_py.run(self)


# }}}

# Patched `develop` command. {{{


class Develop(develop):
    build_dependencies = []

    def run(self):
        for command in self.build_dependencies:
            self.run_command(command)
        develop.run(self)


# }}}



# -*- coding: utf-8 -*-
# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"Launch the plover application."

import argparse
import atexit
import os
import sys
import subprocess
import traceback

from importlib.metadata import entry_points

from plover.config import Config
from plover.oslayer.controller import Controller
from plover.oslayer.config import CONFIG_DIR, CONFIG_FILE, PLATFORM
from plover.registry import registry
from plover import log
from plover import __name__ as __software_name__
from plover import __version__


def init_config_dir():
    """Creates plover's config dir.

    This usually only does anything the first time plover is launched.
    """
    # Create the configuration directory if needed.
    if not os.path.exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)

    # Create a default configuration file if one doesn't already exist.
    if not os.path.exists(CONFIG_FILE):
        open(CONFIG_FILE, "wb").close()


def main():
    """Launch plover."""
    description = "Run the plover stenotype engine. This is a graphical application."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--version",
        action="version",
        version="%s %s" % (__software_name__.capitalize(), __version__),
    )
    parser.add_argument(
        "-s",
        "--script",
        default=None,
        nargs=argparse.REMAINDER,
        help="use another plugin console script as main entrypoint, "
        "passing in the rest of the command line arguments, "
        "print list of available scripts when no argument is given",
    )
    parser.add_argument(
        "-l",
        "--log-level",
        choices=["debug", "info", "warning", "error"],
        default=None,
        help="set log level",
    )
    parser.add_argument("-g", "--gui", default=None, help="set gui")
    args = parser.parse_args(args=sys.argv[1:])
    if args.log_level is not None:
        log.set_level(args.log_level.upper())
    log.setup_platform_handler()

    log.info("Plover %s", __version__)
    log.info("configuration directory: %s", CONFIG_DIR)

    registry.update()

    if args.gui is None:
        gui_priority = {
            "qt": 1,
            "none": -1,
        }
        gui_list = sorted(
            registry.list_plugins("gui"),
            reverse=True,
            key=lambda gui: gui_priority.get(gui.name, 0),
        )
        gui = gui_list[0].obj
    else:
        gui = registry.get_plugin("gui", args.gui).obj

    try:
        if args.script is not None:
            if args.script:
                # Create a mapping of available console scripts,
                # with the following priorities (highest first):
                # - {project_name}-{version}:{script_name}
                # - {project_name}:{script_name}
                # - {script_name}
                console_scripts = {}
                for e in sorted(
                    entry_points(group="console_scripts"),
                    key=lambda e: (e.dist.name, e.name),
                ):
                    for key in (
                        f"{e.dist.name}-{e.dist.version}:{e.name}",
                        f"{e.dist.name}:{e.name}",
                        e.name,
                    ):
                        console_scripts[key] = e
                entrypoint = console_scripts.get(args.script[0])
                if entrypoint is None:
                    log.error("no such script: %s", args.script[0])
                    code = 1
                else:
                    sys.argv = args.script
                    try:
                        code = entrypoint.load()()
                    except SystemExit as e:
                        code = e.code
                if code is None:
                    code = 0
            else:
                print("available script(s):")
                dist = None
                for e in sorted(
                    entry_points(group="console_scripts"),
                    key=lambda e: (e.dist.name, e.name),
                ):
                    if dist != e.dist:
                        dist = e.dist
                        print(f"{dist.name}-{dist.version}:")
                    print(f"- {e.name}")
                code = 0
            os._exit(code)

        # Ensure only one instance of Plover is running at a time.
        with Controller() as controller:
            if controller.is_owner:
                # No other instance, regular startup.
                if PLATFORM == "mac":
                    import appnope

                    appnope.nope()
                init_config_dir()
                # This must be done after calling init_config_dir, so
                # Plover's configuration directory actually exists.
                log.setup_logfile()
                config = Config(CONFIG_FILE)
                code = gui.main(config, controller)
            else:
                log.info("another instance is running, sending `focus` command")
                # Other instance? Try focusing the main window.
                try:
                    controller.send_command("focus")
                except (ConnectionRefusedError, ConnectionResetError):
                    log.error(
                        "connection to existing instance failed, "
                        "force cleaning before restart"
                    )
                    # Assume the previous instance died, leaving
                    # a stray socket, try cleaning it...
                    if not controller.force_cleanup():
                        raise
                    # ...and restart.
                    code = -1
                else:
                    code = 0
    except:
        gui.show_error("Unexpected error", traceback.format_exc())
        code = 2
    # Execute atexit handlers.
    atexit._run_exitfuncs()
    if code == -1:
        # Restart.
        args = sys.argv[:]
        if args[0].endswith(".py") or args[0].endswith(".pyc"):
            # We're running from source.
            spec = sys.modules["__main__"].__spec__
            assert sys.argv[0] == spec.origin
            args[0:1] = [sys.executable, "-m", spec.name]
        if PLATFORM == "win":
            # Workaround https://bugs.python.org/issue19066
            subprocess.Popen(args, cwd=os.getcwd())
            code = 0
        else:
            os.execv(args[0], args)
    os._exit(code)


if __name__ == "__main__":
    main()



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Thread-based monitoring of a Gemini PR stenotype machine."""

import binascii

from plover import log
from plover.machine.base import SerialStenotypeBase


# In the Gemini PR protocol, each packet consists of exactly six bytes
# and the most significant bit (MSB) of every byte is used exclusively
# to indicate whether that byte is the first byte of the packet
# (MSB=1) or one of the remaining five bytes of the packet (MSB=0). As
# such, there are really only seven bits of steno data in each packet
# byte. This is why the STENO_KEY_CHART below is visually presented as
# six rows of seven elements instead of six rows of eight elements.
# fmt: off
STENO_KEY_CHART = ("Fn", "#1", "#2", "#3", "#4", "#5", "#6",
                   "S1-", "S2-", "T-", "K-", "P-", "W-", "H-",
                   "R-", "A-", "O-", "*1", "*2", "res1", "res2",
                   "pwr", "*3", "*4", "-E", "-U", "-F", "-R",
                   "-P", "-B", "-L", "-G", "-T", "-S", "-D",
                   "#7", "#8", "#9", "#A", "#B", "#C", "-Z")
# fmt: on

BYTES_PER_STROKE = 6


class GeminiPr(SerialStenotypeBase):
    """Standard stenotype interface for a Gemini PR machine."""

    KEYS_LAYOUT = """
        #1 #2  #3 #4 #5 #6 #7 #8 #9 #A #B #C
        Fn S1- T- P- H- *1 *3 -F -P -L -T -D
           S2- K- W- R- *2 *4 -R -B -G -S -Z
                  A- O-       -E -U
        pwr
        res1
        res2
    """

    def run(self):
        """Overrides base class run method. Do not call directly."""
        self._ready()
        for packet in self._iter_packets(BYTES_PER_STROKE):
            if not (packet[0] & 0x80) or sum(b & 0x80 for b in packet[1:]):
                log.error("discarding invalid packet: %s", binascii.hexlify(packet))
                continue
            steno_keys = []
            for i, b in enumerate(packet):
                for j in range(1, 8):
                    if b & (0x80 >> j):
                        steno_keys.append(STENO_KEY_CHART[i * 7 + j - 1])
            self._notify_keys(steno_keys)



from contextlib import contextmanager
from importlib.util import find_spec
from tempfile import NamedTemporaryFile
import os
import shutil


ASSET_SCHEME = "asset:"


def _asset_filename(resource_name):
    components = resource_name[len(ASSET_SCHEME) :].split(":", 1)
    if len(components) != 2:
        raise ValueError(f"invalid asset: {resource_name}")
    if os.path.isabs(components[1]):
        raise ValueError(f"invalid asset: {resource_name}")
    # Check for a leading (back)slash since os.path.isabs
    # returns False on Windows if the given path starts with
    # exactly one (back)slash.
    if components[1].startswith(("/", "\\")):
        raise ValueError(f"invalid asset: {resource_name}")
    package_dir = os.path.dirname(find_spec(components[0]).origin)
    return os.path.join(package_dir, components[1])


def resource_exists(resource_name):
    if resource_name.startswith(ASSET_SCHEME):
        resource_name = _asset_filename(resource_name)
    return os.path.exists(resource_name)


def resource_filename(resource_name):
    if resource_name.startswith(ASSET_SCHEME):
        resource_name = _asset_filename(resource_name)
    return resource_name


def resource_timestamp(resource_name):
    filename = resource_filename(resource_name)
    return os.path.getmtime(filename)


@contextmanager
def resource_update(resource_name):
    if resource_name.startswith(ASSET_SCHEME):
        raise ValueError(f"updating an asset is unsupported: {resource_name}")
    filename = resource_filename(resource_name)
    directory = os.path.dirname(filename)
    extension = os.path.splitext(filename)[1]
    tempfile = NamedTemporaryFile(delete=False, dir=directory, suffix=extension or None)
    try:
        tempfile.close()
        yield tempfile.name
        shutil.move(tempfile.name, filename)
    finally:
        if os.path.exists(tempfile.name):
            os.unlink(tempfile.name)



from Xlib import X, display
from Xlib.error import BadWindow
from Xlib.protocol.event import ClientMessage


class WmCtrl:
    def __init__(self):
        self._display = display.Display()
        self._root = self._display.screen().root
        self._atoms = {
            name: self._display.intern_atom(name)
            for name in """
            _NET_ACTIVE_WINDOW
            _NET_CURRENT_DESKTOP
            _NET_WM_DESKTOP
            _WIN_WORKSPACE
            """.split()
        }

    def _get_wm_property(self, window, atom_name):
        prop = window.get_full_property(self._atoms[atom_name], X.AnyPropertyType)
        return None if prop is None else prop.value[0]

    def _client_msg(self, window, atom_name, data):
        ev_data = (data,) + (0,) * 4
        ev = ClientMessage(
            window=window, client_type=self._atoms[atom_name], data=(32, ev_data)
        )
        ev_mask = X.SubstructureRedirectMask | X.SubstructureNotifyMask
        self._root.send_event(ev, event_mask=ev_mask)
        self._display.sync()

    def _map_raised(self, window):
        window.map()
        window.raise_window()
        self._display.sync()

    def get_foreground_window(self):
        return self._get_wm_property(self._root, "_NET_ACTIVE_WINDOW")

    def set_foreground_window(self, w):
        try:
            window = self._display.create_resource_object("window", w)
            for atom in ("_NET_WM_DESKTOP", "_WIN_WORKSPACE"):
                desktop = self._get_wm_property(window, atom)
                if desktop is not None:
                    self._client_msg(self._root, "_NET_CURRENT_DESKTOP", desktop)
                    break
            self._client_msg(window, "_NET_ACTIVE_WINDOW", 0)
            self._map_raised(window)
        except BadWindow:
            pass



from contextlib import contextmanager
import ctypes.util
import os
import logging

from plover import log, __name__ as __software_name__
from plover.oslayer.config import ASSETS_DIR


APPNAME = ctypes.c_char_p(__software_name__.capitalize().encode())
APPICON = ctypes.c_char_p(os.path.join(ASSETS_DIR, "plover.png").encode())
SERVICE = ctypes.c_char_p(b"org.freedesktop.Notifications")
INTERFACE = ctypes.c_char_p(b"/org/freedesktop/Notifications")

NOTIFY_URGENCY_LOW = ctypes.c_uint8(0)
NOTIFY_URGENCY_NORMAL = ctypes.c_uint8(1)
NOTIFY_URGENCY_CRITICAL = ctypes.c_uint8(2)

DBUS_BUS_SESSION = ctypes.c_uint(0)

DBUS_TYPE_ARRAY = ctypes.c_int(ord("a"))
DBUS_TYPE_BYTE = ctypes.c_int(ord("y"))
DBUS_TYPE_DICT_ENTRY = ctypes.c_int(ord("e"))
DBUS_TYPE_INT32 = ctypes.c_int(ord("i"))
DBUS_TYPE_STRING = ctypes.c_int(ord("s"))
DBUS_TYPE_UINT32 = ctypes.c_int(ord("u"))
DBUS_TYPE_VARIANT = ctypes.c_int(ord("v"))


class DBusConnection(ctypes.c_void_p):
    pass


class DBusError(ctypes.Structure):
    _fields_ = (
        ("name", ctypes.c_char_p),
        ("message", ctypes.c_char_p),
        ("dummy1", ctypes.c_uint),
        ("dummy2", ctypes.c_uint),
        ("dummy3", ctypes.c_uint),
        ("dummy4", ctypes.c_uint),
        ("dummy5", ctypes.c_uint),
        ("padding1", ctypes.c_void_p),
    )


class DBusMessage(ctypes.c_void_p):
    pass


class DBusMessageIter(ctypes.Structure):
    _fields_ = (
        ("dummy1", ctypes.c_void_p),
        ("dummy2", ctypes.c_void_p),
        ("dummy3", ctypes.c_uint32),
        ("dummy4", ctypes.c_int),
        ("dummy5", ctypes.c_int),
        ("dummy6", ctypes.c_int),
        ("dummy7", ctypes.c_int),
        ("dummy8", ctypes.c_int),
        ("dummy9", ctypes.c_int),
        ("dummy10", ctypes.c_int),
        ("dummy11", ctypes.c_int),
        ("pad1", ctypes.c_int),
        ("pad2", ctypes.c_void_p),
        ("pad3", ctypes.c_void_p),
    )


def ctypes_type(signature):
    if signature == "void":
        return None
    ct = {
        "connection_p": DBusConnection,
        "error_p": DBusError,
        "message_p": DBusMessage,
        "message_iter_p": DBusMessageIter,
    }.get(signature)
    if ct is not None:
        return ctypes.POINTER(ct)
    ct = getattr(ctypes, "c_" + signature, None)
    if ct is not None:
        return ct
    if not signature.endswith("_p"):
        raise ValueError(signature)
    ct = getattr(ctypes, "c_" + signature[:-2])
    if ct is None:
        raise ValueError(signature)
    return ctypes.POINTER(ct)


def ctypes_func(library, signature):
    restype, func_name, *argtypes = signature.split()
    func = getattr(library, func_name)
    func.argtypes = tuple(map(ctypes_type, argtypes))
    func.restype = ctypes_type(restype)
    return func


class DBusNotificationHandler(logging.Handler):
    """Handler using DBus notifications to show messages."""

    def __init__(self):
        super().__init__()
        self.setLevel(log.WARNING)
        self.setFormatter(
            log.NoExceptionTracebackFormatter("<b>%(levelname)s:</b> %(message)s")
        )

        libname = ctypes.util.find_library("dbus-1")
        if libname is None:
            raise FileNotFoundError("dbus-1 library")
        library = ctypes.cdll.LoadLibrary(libname)

        error_free = ctypes_func(library, "void dbus_error_free error_p")
        error_init = ctypes_func(library, "void dbus_error_init error_p")
        error_is_set = ctypes_func(library, "bool dbus_error_is_set error_p")
        bus_get = ctypes_func(library, "connection_p dbus_bus_get uint error_p")
        message_new = ctypes_func(
            library,
            "message_p dbus_message_new_method_call char_p char_p char_p char_p",
        )
        message_unref = ctypes_func(library, "void dbus_message_unref message_p")
        iter_init_append = ctypes_func(
            library, "void dbus_message_iter_init_append message_p message_iter_p"
        )
        iter_append_basic = ctypes_func(
            library, "bool dbus_message_iter_append_basic message_iter_p int void_p"
        )
        iter_open_container = ctypes_func(
            library,
            "bool dbus_message_iter_open_container message_iter_p int char_p message_iter_p",
        )
        iter_close_container = ctypes_func(
            library,
            "bool dbus_message_iter_close_container message_iter_p message_iter_p",
        )
        connection_send = ctypes_func(
            library, "bool dbus_connection_send connection_p message_p uint32_p"
        )

        # Need message + container + dict_entry + variant = 4 iterators.
        self._iter_stack = [DBusMessageIter() for __ in range(4)]
        self._iter_stack_index = 0

        @contextmanager
        def open_container(kind, signature):
            parent_iter = self._iter_stack[self._iter_stack_index]
            sub_iter = self._iter_stack[self._iter_stack_index + 1]
            if not iter_open_container(parent_iter, kind, signature, sub_iter):
                raise MemoryError
            self._iter_stack_index += 1
            try:
                yield
            finally:
                if not iter_close_container(parent_iter, sub_iter):
                    raise MemoryError
                self._iter_stack_index -= 1

        def append_basic(kind, value):
            if not iter_append_basic(
                self._iter_stack[self._iter_stack_index], kind, ctypes.byref(value)
            ):
                raise MemoryError

        error = DBusError()
        error_init(error)
        bus = bus_get(DBUS_BUS_SESSION, ctypes.byref(error))
        if error_is_set(error):
            e = ConnectionError(
                "%s: %s" % (error.name.decode(), error.message.decode())
            )
            error_free(error)
            raise e
        assert bus is not None

        actions_signature = ctypes.c_char_p(b"s")
        hints_signature = ctypes.c_char_p(b"{sv}")
        notify_str = ctypes.c_char_p(b"Notify")
        urgency_signature = ctypes.c_char_p(b"y")
        urgency_str = ctypes.c_char_p(b"urgency")
        zero = ctypes.c_uint(0)

        def notify(body, urgency, timeout):
            message = message_new(SERVICE, INTERFACE, SERVICE, notify_str)
            try:
                iter_init_append(message, self._iter_stack[self._iter_stack_index])
                # app_name
                append_basic(DBUS_TYPE_STRING, APPNAME)
                # replaces_id
                append_basic(DBUS_TYPE_UINT32, zero)
                # app_icon
                append_basic(DBUS_TYPE_STRING, APPICON)
                # summary
                append_basic(DBUS_TYPE_STRING, APPNAME)
                # body
                append_basic(DBUS_TYPE_STRING, body)
                # actions
                with open_container(DBUS_TYPE_ARRAY, actions_signature):
                    pass
                # hints
                with (
                    open_container(DBUS_TYPE_ARRAY, hints_signature),
                    open_container(DBUS_TYPE_DICT_ENTRY, None),
                ):
                    append_basic(DBUS_TYPE_STRING, urgency_str)
                    with open_container(DBUS_TYPE_VARIANT, urgency_signature):
                        append_basic(DBUS_TYPE_BYTE, urgency)
                # expire_timeout
                append_basic(DBUS_TYPE_INT32, timeout)
                connection_send(bus, message, None)
            finally:
                message_unref(message)

        self._notify = notify

    def handle(self, record):
        level = record.levelno
        message = self.format(record)
        if message.endswith("\n"):
            message = message[:-1]
        if level <= log.INFO:
            timeout = 10
            urgency = NOTIFY_URGENCY_LOW
        elif level <= log.WARNING:
            timeout = 15
            urgency = NOTIFY_URGENCY_NORMAL
        else:
            timeout = 0
            urgency = NOTIFY_URGENCY_CRITICAL
        self._notify(
            ctypes.c_char_p(message.encode()), urgency, ctypes.c_int(timeout * 1000)
        )



from unittest.mock import Mock
from plover.machine.base import ThreadedStenotypeBase
import pytest


class MyMachine(ThreadedStenotypeBase):
    def run(self):
        raise RuntimeError("some unexpected error")


@pytest.mark.filterwarnings("ignore::pytest.PytestUnhandledThreadExceptionWarning")
def test_update_machine_staten_on_unhandled_exception():
    machine = MyMachine()
    callback = Mock()
    machine.add_state_callback(callback)
    machine.start_capture()
    machine.join()
    callback.assert_called_with("disconnected")



from functools import partial
import json
import os
import subprocess

from PySide6.QtCore import QCoreApplication, Qt, Slot
from PySide6.QtGui import QCursor, QIcon, QKeySequence
from PySide6.QtWidgets import (
    QMainWindow,
    QMenu,
)

from plover import _, log
from plover.oslayer import wmctrl
from plover.oslayer.config import CONFIG_DIR, PLATFORM
from plover.registry import registry

from plover.gui_qt import utils
from plover.gui_qt.log_qt import NotificationHandler
from plover.gui_qt.main_window_ui import Ui_MainWindow
from plover.gui_qt.config_window import ConfigWindow
from plover.gui_qt.about_dialog import AboutDialog
from plover.gui_qt.trayicon import TrayIcon
from plover.gui_qt.utils import WindowStateMixin


class MainWindow(QMainWindow, Ui_MainWindow, WindowStateMixin):
    ROLE = "main"

    def __init__(self, engine, use_qt_notifications):
        super().__init__()
        self.setupUi(self)
        if hasattr(self, "setUnifiedTitleAndToolBarOnMac"):
            self.setUnifiedTitleAndToolBarOnMac(True)
        self._engine = engine
        self._active_dialogs = {}
        self._dialog_class = {
            "about": AboutDialog,
            "configuration": ConfigWindow,
        }
        all_actions = utils.find_menu_actions(self.menubar)
        # Dictionaries.
        self.dictionaries.signal_add_translation.connect(self._add_translation)
        self.dictionaries.setup(engine)
        # Populate edit menu from dictionaries' own.
        edit_menu = all_actions["menu_Edit"].menu()
        for action in self.dictionaries.edit_menu.actions():
            edit_menu.addAction(action)
        # Tray icon.
        self._trayicon = TrayIcon()
        self._trayicon.enable()
        self._trayicon.clicked.connect(self._engine.toggle_output)
        if use_qt_notifications:
            handler = NotificationHandler()
            handler.emitSignal.connect(self._trayicon.log)
            log.add_handler(handler)
        popup_menu = QMenu()
        for action_name in (
            "action_ToggleOutput",
            "action_Reconnect",
            "",
            "menu_Tools",
            "",
            "action_Configure",
            "action_OpenConfigFolder",
            "",
            "menu_Help",
            "",
            "action_Show",
            "action_Quit",
        ):
            if action_name:
                popup_menu.addAction(all_actions[action_name])
            else:
                popup_menu.addSeparator()
        self._trayicon.set_menu(popup_menu)
        engine.signal_connect(
            "machine_state_changed", self._trayicon.update_machine_state
        )
        engine.signal_connect("quit", self.quit_application)
        self.action_Quit.triggered.connect(engine.quit)
        # Toolbar popup menu for selecting which tools are shown.
        self.toolbar_menu = QMenu()
        self.toolbar.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.toolbar.customContextMenuRequested.connect(
            lambda: self.toolbar_menu.popup(QCursor.pos())
        )
        # Populate tools bar/menu.
        tools_menu = all_actions["menu_Tools"].menu()
        for tool_plugin in registry.list_plugins("gui.qt.tool"):
            tool = tool_plugin.obj
            menu_action = tools_menu.addAction(tool.TITLE)
            if tool.SHORTCUT is not None:
                menu_action.setShortcut(QKeySequence.fromString(tool.SHORTCUT))
            if tool.ICON is not None:
                menu_action.setIcon(QIcon(tool.ICON))
            menu_action.triggered.connect(
                partial(self._activate_dialog, tool_plugin.name, args=())
            )
            toolbar_action = self.toolbar.addAction(
                menu_action.icon(), menu_action.text()
            )
            if tool.__doc__ is not None:
                toolbar_action.setToolTip(tool.__doc__)
            toolbar_action.triggered.connect(menu_action.trigger)
            toggle_action = self.toolbar_menu.addAction(
                menu_action.icon(), menu_action.text()
            )
            toggle_action.setObjectName(tool_plugin.name)
            toggle_action.setCheckable(True)
            toggle_action.setChecked(True)
            toggle_action.toggled.connect(toolbar_action.setVisible)
            self._dialog_class[tool_plugin.name] = tool
        engine.signal_connect("output_changed", self.initialize_output_state)
        # Machine.
        for plugin in registry.list_plugins("machine"):
            self.machine_type.addItem(_(plugin.name), plugin.name)
        engine.signal_connect("config_changed", self.on_config_changed)
        engine.signal_connect(
            "machine_state_changed",
            lambda machine, state: self.machine_state.setText(state.capitalize()),
        )
        self.restore_state()
        # Commands.
        engine.signal_connect(
            "add_translation", partial(self._add_translation, manage_windows=True)
        )
        engine.signal_connect("focus", self._focus)
        engine.signal_connect(
            "configure", partial(self._configure, manage_windows=True)
        )
        engine.signal_connect(
            "lookup", partial(self._activate_dialog, "lookup", manage_windows=True)
        )
        engine.signal_connect(
            "suggestions",
            partial(self._activate_dialog, "suggestions", manage_windows=True),
        )
        # Load the configuration (but do not start the engine yet).
        if not engine.load_config():
            self.configure()
        # Apply configuration settings.
        config = self._engine.config
        self._warn_on_hide_to_tray = not config["start_minimized"]
        self._update_machine(config["machine_type"])
        self._configured = False
        self.set_visible(not config["start_minimized"])
        # Process events before starting the engine
        # (to avoid display lag at window creation).
        QCoreApplication.processEvents()
        # Start the engine.
        engine.start()

    def set_visible(self, visible):
        if visible:
            self.show()
        else:
            if self._trayicon.is_enabled():
                self.hide()
            else:
                self.showMinimized()

    def _activate_dialog(self, name, args=(), manage_windows=False):
        if manage_windows:
            previous_window = wmctrl.GetForegroundWindow()
        dialog = self._active_dialogs.get(name)
        if dialog is None:
            dialog_class = self._dialog_class[name]
            dialog = self._active_dialogs[name] = dialog_class(self._engine, *args)
            dialog.setWindowIcon(self.windowIcon())

            def on_finished():
                del self._active_dialogs[name]
                dialog.deleteLater()
                if manage_windows and previous_window is not None:
                    wmctrl.SetForegroundWindow(previous_window)

            dialog.finished.connect(on_finished)
        dialog.showNormal()
        if not utils.is_wayland():
            # Otherwise gives this warning:
            # Qt: Wayland does not support QWindow::requestActivate()
            dialog.activateWindow()
        dialog.raise_()

    def _add_translation(self, dictionary=None, manage_windows=False):
        if not dictionary:
            dictionary = None
        self._activate_dialog(
            "add_translation", args=(dictionary,), manage_windows=manage_windows
        )

    def _focus(self):
        self.showNormal()
        if not utils.is_wayland():
            self.activateWindow()
        self.raise_()

    def _configure(self, manage_windows=False):
        self._activate_dialog("configuration", manage_windows=manage_windows)

    def _lookup(self, manage_windows=False):
        self._activate_dialog("lookup", manage_windows=manage_windows)

    def _restore_state(self, settings):
        if settings.contains("hidden_toolbar_tools"):
            hidden_toolbar_tools = json.loads(settings.value("hidden_toolbar_tools"))
            for action in self.toolbar_menu.actions():
                action.setChecked(action.objectName() not in hidden_toolbar_tools)

    def _save_state(self, settings):
        hidden_toolbar_tools = {
            action.objectName()
            for action in self.toolbar_menu.actions()
            if not action.isChecked()
        }
        settings.setValue(
            "hidden_toolbar_tools", json.dumps(list(sorted(hidden_toolbar_tools)))
        )

    def _update_machine(self, machine_type):
        self.machine_type.setCurrentIndex(self.machine_type.findData(machine_type))

    def on_config_changed(self, config_update):
        if "machine_type" in config_update:
            self._update_machine(config_update["machine_type"])
        if not self._configured:
            self._configured = True
            if config_update.get("show_suggestions_display", False):
                self._activate_dialog("suggestions")
            if config_update.get("show_stroke_display", False):
                self._activate_dialog("paper_tape")

    @Slot(int)
    def update_machine_type(self, machine_index):
        self._engine.config = {
            "machine_type": self.machine_type.itemData(machine_index)
        }

    @Slot(bool)
    def initialize_output_state(self, enabled):
        self._trayicon.update_output(enabled)
        self.output_enable.setChecked(enabled)
        self.output_disable.setChecked(not enabled)
        self.action_ToggleOutput.setChecked(enabled)

    @Slot(bool)
    def toggle_output(self, enabled):
        self._engine.output = enabled

    @Slot()
    def enable_output(self):
        self.toggle_output(True)

    @Slot()
    def disable_output(self):
        self.toggle_output(False)

    @Slot()
    def configure(self):
        self._configure()

    @Slot()
    def open_config_folder(self):
        if PLATFORM == "win":
            os.startfile(CONFIG_DIR)
        elif PLATFORM == "linux":
            subprocess.call(["xdg-open", CONFIG_DIR])
        elif PLATFORM == "mac":
            subprocess.call(["open", CONFIG_DIR])

    @Slot()
    def reconnect(self):
        self._engine.reset_machine()

    @Slot()
    def open_about_dialog(self):
        self._activate_dialog("about")

    @Slot()
    def quit_application(self):
        for dialog in list(self._active_dialogs.values()):
            dialog.close()
        self.save_state()
        self._trayicon.disable()
        self.hide()
        QCoreApplication.quit()

    def show_window(self):
        self._focus()

    def closeEvent(self, event):
        event.ignore()
        self.hide()
        if not self._trayicon.is_enabled():
            self._engine.quit()
            return
        if not self._warn_on_hide_to_tray:
            return
        self._trayicon.show_message(_("Application is still running."))
        self._warn_on_hide_to_tray = False



# Copyright (c) 2011 Hesky Fisher.
# See LICENSE.txt for details.
#
# winkeyboardcontrol.py - capturing and injecting keyboard events in windows.

"""Keyboard capture and control in windows.

This module provides an interface for basic keyboard event capture and
emulation. Set the key_up and key_down functions of the
KeyboardCapture class to capture keyboard input. Call the send_string
and send_backspaces functions of the KeyboardEmulation class to
emulate keyboard input.

"""

from ctypes import windll, wintypes
import atexit
import ctypes
import multiprocessing
import os
import signal
import threading
import winreg

from plover import log
from plover.key_combo import parse_key_combo
from plover.machine.keyboard_capture import Capture
from plover.misc import to_surrogate_pair
from plover.output.keyboard import GenericKeyboardEmulation

from .keyboardlayout import KeyboardLayout

# For the purposes of this class, we'll only report key presses that
# result in these outputs in order to exclude special key combos.
SCANCODE_TO_KEY = {
    59: "F1",
    60: "F2",
    61: "F3",
    62: "F4",
    63: "F5",
    64: "F6",
    65: "F7",
    66: "F8",
    67: "F9",
    68: "F10",
    87: "F11",
    88: "F12",
    41: "`",
    2: "1",
    3: "2",
    4: "3",
    5: "4",
    6: "5",
    7: "6",
    8: "7",
    9: "8",
    10: "9",
    11: "0",
    12: "-",
    13: "=",
    16: "q",
    17: "w",
    18: "e",
    19: "r",
    20: "t",
    21: "y",
    22: "u",
    23: "i",
    24: "o",
    25: "p",
    26: "[",
    27: "]",
    43: "\\",
    30: "a",
    31: "s",
    32: "d",
    33: "f",
    34: "g",
    35: "h",
    36: "j",
    37: "k",
    38: "l",
    39: ";",
    40: "'",
    44: "z",
    45: "x",
    46: "c",
    47: "v",
    48: "b",
    49: "n",
    50: "m",
    51: ",",
    52: ".",
    53: "/",
    57: "space",
    14: "BackSpace",
    83: "Delete",
    80: "Down",
    79: "End",
    1: "Escape",
    71: "Home",
    82: "Insert",
    75: "Left",
    73: "Page_Up",
    81: "Page_Down",
    28: "Return",
    77: "Right",
    15: "Tab",
    72: "Up",
}

KEY_TO_SCANCODE = dict(zip(SCANCODE_TO_KEY.values(), SCANCODE_TO_KEY.keys()))

PASSTHROUGH_KEYS = {
    0xA2,  # Control
    0xA3,  # Control
    0xA0,  # Shift
    0xA1,  # Shift
    0xA4,  # Alt
    0xA5,  # Alt
    0x5B,  # Win
    0x5C,  # Win
}


"""
SendInput code and classes based off:
http://stackoverflow.com/questions/11906925/python-simulate-keydown
"""


class MOUSEINPUT(ctypes.Structure):
    _fields_ = (
        ("dx", wintypes.LONG),
        ("dy", wintypes.LONG),
        ("mouseData", wintypes.DWORD),
        ("dwFlags", wintypes.DWORD),
        ("time", wintypes.DWORD),
        ("dwExtraInfo", wintypes.PULONG),
    )


class KEYBDINPUT(ctypes.Structure):
    _fields_ = (
        ("wVk", wintypes.WORD),
        ("wScan", wintypes.WORD),
        ("dwFlags", wintypes.DWORD),
        ("time", wintypes.DWORD),
        ("dwExtraInfo", wintypes.PULONG),
    )


class _INPUTunion(ctypes.Union):
    _fields_ = (("mi", MOUSEINPUT), ("ki", KEYBDINPUT))


class INPUT(ctypes.Structure):
    _fields_ = (("type", wintypes.DWORD), ("union", _INPUTunion))


SendInput = windll.user32.SendInput
SendInput.argtypes = [
    wintypes.UINT,  # cInputs
    ctypes.POINTER(INPUT),  # pInputs,
    ctypes.c_int,  # cbSize
]
SendInput.restype = wintypes.UINT

KEYEVENTF_EXTENDEDKEY = 0x0001
KEYEVENTF_KEYUP = 0x0002
KEYEVENTF_SCANCODE = 0x0008
KEYEVENTF_UNICODE = 0x0004

INPUT_MOUSE = 0
INPUT_KEYBOARD = 1

KF_EXTENDED = 0x0100
LLKHF_EXTENDED = KF_EXTENDED >> 8
LLKHF_LOWER_IL_INJECTED = 0x0002
LLKHF_INJECTED = 0x0010


# Process utils code code based on psutil implementation.

OpenProcess = windll.kernel32.OpenProcess
OpenProcess.argtypes = [
    wintypes.DWORD,  # dwDesiredAccess
    wintypes.BOOL,  # bInheritHandle
    wintypes.DWORD,  # dwProcessId
]
OpenProcess.restype = wintypes.HANDLE

GetExitCodeProcess = windll.kernel32.GetExitCodeProcess
GetExitCodeProcess.argtypes = [
    wintypes.HANDLE,  # hProcess
    wintypes.LPDWORD,  # lpExitCode
]
GetExitCodeProcess.restype = wintypes.BOOL

CloseHandle = windll.kernel32.CloseHandle
CloseHandle.argtypes = [
    wintypes.HANDLE,  # hObject
]
CloseHandle.restype = wintypes.BOOL

PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_VM_READ = 0x0010
ERROR_INVALID_PARAMETER = 87
ERROR_ACCESS_DENIED = 5
STILL_ACTIVE = 0x00000103


def pid_exists(pid):
    """Check whether pid exists in the current process table."""

    process = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, 0, pid)
    if not process:
        err = ctypes.GetLastError()
        if err == ERROR_INVALID_PARAMETER:
            # Invalid parameter is no such process.
            return False
        if err == ERROR_ACCESS_DENIED:
            # Access denied obviously means there's a process to deny access to...
            return True
        raise ctypes.WinError(err)

    try:
        exitcode = wintypes.DWORD()
        out = GetExitCodeProcess(process, ctypes.byref(exitcode))
        if not out:
            err = ctypes.GetLastError()
            if err == ERROR_ACCESS_DENIED:
                # Access denied means there's a process
                # there so we'll assume it's running.
                return True
            raise ctypes.WinError(err)

        return exitcode.value == STILL_ACTIVE

    finally:
        CloseHandle(process)


class HeartBeat(threading.Thread):
    def __init__(self, ppid, exitcb):
        super().__init__()
        self._ppid = ppid
        self._exitcb = exitcb
        self._finished = threading.Event()

    def run(self):
        while pid_exists(self._ppid):
            if self._finished.wait(1):
                break
        self._exitcb()

    def stop(self):
        self._finished.set()
        self.join()


class KBDLLHOOKSTRUCT(ctypes.Structure):
    _fields_ = (
        ("vkCode", wintypes.DWORD),
        ("scanCode", wintypes.DWORD),
        ("flags", wintypes.DWORD),
        ("time", wintypes.DWORD),
        ("dwExtraInfo", ctypes.c_void_p),
    )


PKBDLLHOOKSTRUCT = ctypes.POINTER(KBDLLHOOKSTRUCT)
LRESULT = ctypes.c_long
HOOKPROC = ctypes.CFUNCTYPE(LRESULT, ctypes.c_int, wintypes.WPARAM, wintypes.LPARAM)

SetWindowsHookExA = windll.user32.SetWindowsHookExA
SetWindowsHookExA.argtypes = (
    ctypes.c_int,  # idHook,
    HOOKPROC,  # lpfn,
    wintypes.HINSTANCE,  # hmod,
    wintypes.DWORD,  # dwThreadId
)
SetWindowsHookExA.restype = wintypes.HHOOK

CallNextHookEx = windll.user32.CallNextHookEx
CallNextHookEx.argtypes = (
    wintypes.HHOOK,  # hhk
    ctypes.c_int,  # nCode
    wintypes.WPARAM,  # wParam
    wintypes.LPARAM,  # lParam
)
CallNextHookEx.restype = LRESULT

UnhookWindowsHookEx = windll.user32.UnhookWindowsHookEx
UnhookWindowsHookEx.argtypes = (wintypes.HHOOK,)
UnhookWindowsHookEx.restype = wintypes.BOOL

GetMessageW = windll.user32.GetMessageW
GetMessageW.argtypes = (
    wintypes.LPMSG,  # lpMsg,
    wintypes.HWND,  # hWnd,
    wintypes.UINT,  # wMsgFilterMin,
    wintypes.UINT,  # wMsgFilterMax
)
GetMessageW.restype = wintypes.BOOL

TranslateMessage = windll.user32.TranslateMessage
TranslateMessage.argtypes = (wintypes.MSG,)
TranslateMessage.restype = wintypes.BOOL

DispatchMessageW = windll.user32.DispatchMessageW
DispatchMessageW.argtypes = (wintypes.MSG,)
DispatchMessageW.restype = LRESULT

PostThreadMessageW = windll.user32.PostThreadMessageW
PostThreadMessageW.argtypes = (
    wintypes.DWORD,  # idThread
    wintypes.UINT,  # Msg
    wintypes.WPARAM,  # wParam
    wintypes.LPARAM,  # lParam
)
PostThreadMessageW.restype = wintypes.BOOL

WM_QUIT = 0x12

# Registry setting for low level hook timeout.
REG_LLHOOK_KEY_FULL_NAME = (
    r"HKEY_CURRENT_USER\Control Panel\Desktop\LowLevelHooksTimeout"
)
REG_LLHOOK_KEY_VALUE_NAME = "LowLevelHooksTimeout"
REG_LLHOOK_KEY_VALUE_TYPE = winreg.REG_DWORD
REG_LLHOOK_KEY_VALUE = 5000


class KeyboardCaptureProcess(multiprocessing.Process):
    def __init__(self):
        super().__init__()
        self.daemon = True
        self._ppid = os.getpid()
        self._update_registry()
        self._tid = None
        self._queue = multiprocessing.Queue()
        self._suppressed_keys_bitmask = multiprocessing.Array(
            ctypes.c_uint64, (max(SCANCODE_TO_KEY.keys()) + 63) // 64
        )
        self._suppressed_keys_bitmask[:] = (0xFFFFFFFFFFFFFFFF,) * len(
            self._suppressed_keys_bitmask
        )

    @staticmethod
    def _update_registry():
        # From MSDN documentation:
        #
        # The hook procedure should process a message in less time than the
        # data entry specified in the LowLevelHooksTimeout value in the
        # following registry key:
        #
        # HKEY_CURRENT_USER\Control Panel\Desktop
        #
        # The value is in milliseconds. If the hook procedure times out, the
        # system passes the message to the next hook. However, on Windows 7 and
        # later, the hook is silently removed without being called. There is no
        # way for the application to know whether the hook is removed.

        def _open_key(rights):
            return winreg.OpenKey(
                winreg.HKEY_CURRENT_USER, r"Control Panel\Desktop", 0, rights
            )

        read_key = _open_key(winreg.KEY_READ)
        try:
            value, value_type = winreg.QueryValueEx(read_key, REG_LLHOOK_KEY_VALUE_NAME)
        except OSError:
            value, value_type = (None, None)

        if value_type != REG_LLHOOK_KEY_VALUE_TYPE or value != REG_LLHOOK_KEY_VALUE:
            try:
                write_key = _open_key(winreg.KEY_WRITE)
                winreg.SetValueEx(
                    write_key,
                    REG_LLHOOK_KEY_VALUE_NAME,
                    0,
                    REG_LLHOOK_KEY_VALUE_TYPE,
                    REG_LLHOOK_KEY_VALUE,
                )
            except OSError:
                log.warning(
                    "could not update registry key: %s, see documentation",
                    REG_LLHOOK_KEY_FULL_NAME,
                )
            else:
                log.warning(
                    "the following registry key has been updated, "
                    "you should reboot: %s",
                    REG_LLHOOK_KEY_FULL_NAME,
                )

    def run(self):
        heartbeat = HeartBeat(self._ppid, self._send_quit)
        heartbeat.start()
        try:
            self._run()
        finally:
            heartbeat.stop()

    def _run(self):
        # Ignore KeyboardInterrupt when attached to a console...
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        self._tid = windll.kernel32.GetCurrentThreadId()
        self._queue.put(self._tid)

        passthrough_down_keys = set()

        def on_key(pressed, event):
            if event.flags & LLKHF_EXTENDED:
                return False

            if event.flags & LLKHF_INJECTED:
                # Ignore simulated events (e.g. from KeyboardEmulation).
                return False
            if event.vkCode in PASSTHROUGH_KEYS:
                if pressed:
                    passthrough_down_keys.add(event.vkCode)
                else:
                    passthrough_down_keys.discard(event.vkCode)
            key = SCANCODE_TO_KEY.get(event.scanCode)
            if key is None:
                # Unhandled, ignore and don't suppress.
                return False
            suppressed = bool(
                self._suppressed_keys_bitmask[event.scanCode // 64]
                & (1 << (event.scanCode % 64))
            )
            if pressed and passthrough_down_keys:
                # Modifier(s) pressed, ignore.
                return False
            self._queue.put((None, key, pressed))
            return suppressed

        hook_id = None

        def low_level_handler(code, wparam, lparam):
            if code >= 0:
                event = ctypes.cast(lparam, PKBDLLHOOKSTRUCT)[0]
                pressed = wparam in (0x100, 0x104)
                if on_key(pressed, event):
                    # Suppressed...
                    return 1
            return CallNextHookEx(hook_id, code, wparam, lparam)

        hook_proc = HOOKPROC(low_level_handler)
        hook_id = SetWindowsHookExA(0xD, hook_proc, None, 0)
        if not hook_id:
            self._queue.put(
                (
                    ("failed to install keyboard hook: %s", ctypes.FormatError()),
                    None,
                    None,
                )
            )
            return
        atexit.register(UnhookWindowsHookEx, hook_id)
        msg = wintypes.MSG()
        msg_p = ctypes.byref(msg)
        while GetMessageW(msg_p, None, 0, 0):
            TranslateMessage(msg_p)
            DispatchMessageW(msg_p)

    def _send_quit(self):
        PostThreadMessageW(self._tid, WM_QUIT, 0, 0)

    def start(self):
        self.daemon = True
        super().start()
        self._tid = self._queue.get()

    def stop(self):
        if self.is_alive():
            self._send_quit()
            self.join()
        # Wake up capture thread, so it gets a chance to check if it must stop.
        self._queue.put((None, None, None))

    def suppress(self, suppressed_keys):
        bitmask = [0] * len(self._suppressed_keys_bitmask)
        for key in suppressed_keys:
            code = KEY_TO_SCANCODE[key]
            bitmask[code // 64] |= 1 << (code % 64)
        self._suppressed_keys_bitmask[:] = bitmask

    def get(self):
        return self._queue.get()


class KeyboardCapture(Capture):
    def __init__(self):
        super().__init__()
        self._suppressed_keys = set()
        self._finished = None
        self._thread = None
        self._proc = None

    def start(self):
        self._finished = threading.Event()
        self._proc = KeyboardCaptureProcess()
        self._proc.start()
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def _run(self):
        while True:
            error, key, pressed = self._proc.get()
            if error is not None:
                log.error(*error)
            if self._finished.is_set():
                break
            (self.key_down if pressed else self.key_up)(key)

    def cancel(self):
        if self._finished is not None:
            self._finished.set()
        if self._proc is not None:
            self._proc.stop()
        if self._thread is not None:
            self._thread.join()

    def suppress(self, suppressed_keys=()):
        self._suppressed_keys = set(suppressed_keys)
        self._proc.suppress(self._suppressed_keys)


class KeyboardEmulation(GenericKeyboardEmulation):
    def __init__(self):
        super().__init__()
        self.keyboard_layout = KeyboardLayout()

    # Sends input types to buffer
    @staticmethod
    def _send_input(*inputs):
        len_inputs = len(inputs)
        len_pinput = INPUT * len_inputs
        pinputs = len_pinput(*inputs)
        c_size = ctypes.c_int(ctypes.sizeof(INPUT))
        return SendInput(len_inputs, pinputs, c_size)

    # Input type (can be mouse, keyboard)
    @staticmethod
    def _input(structure):
        if isinstance(structure, MOUSEINPUT):
            return INPUT(INPUT_MOUSE, _INPUTunion(mi=structure))
        if isinstance(structure, KEYBDINPUT):
            return INPUT(INPUT_KEYBOARD, _INPUTunion(ki=structure))
        raise TypeError("Cannot create INPUT structure!")

    # Container to send mouse input
    # Not used, but maybe one day it will be useful
    @staticmethod
    def _mouse_input(flags, x, y, data):
        return MOUSEINPUT(x, y, data, flags, 0, None)

    # Keyboard input type to send key input
    @staticmethod
    def _keyboard_input(code, flags):
        if flags & KEYEVENTF_UNICODE:
            # special handling of Unicode characters
            return KEYBDINPUT(0, code, flags, 0, None)
        return KEYBDINPUT(code, 0, flags, 0, None)

    # Abstraction to set flags to 0 and create an input type
    def _keyboard(self, code, flags=0):
        return self._input(self._keyboard_input(code, flags))

    def _key_event(self, keycode, pressed):
        flags = 0 if pressed else KEYEVENTF_KEYUP
        self._send_input(self._keyboard(keycode, flags))

    # Press and release a key
    def _key_press(self, char):
        vk, ss = self.keyboard_layout.char_to_vk_ss[char]
        keycode_list = []
        keycode_list.extend(self.keyboard_layout.ss_to_vks[ss])
        keycode_list.append(vk)
        # Press all keys.
        for keycode in keycode_list:
            self._key_event(keycode, True)
        # Release all keys
        for keycode in keycode_list:
            self._key_event(keycode, False)

    def _refresh_keyboard_layout(self):
        layout_id = KeyboardLayout.current_layout_id()
        if layout_id != self.keyboard_layout.layout_id:
            self.keyboard_layout = KeyboardLayout(layout_id)

    def _key_unicode(self, char):
        pairs = to_surrogate_pair(char)
        # Send press events for all codes, then release events for all codes.
        inputs = [
            self._keyboard(code, KEYEVENTF_UNICODE | direction)
            for direction in (0, KEYEVENTF_KEYUP)
            for code in pairs
        ]
        self._send_input(*inputs)

    def send_backspaces(self, count):
        for _ in self.with_delay(range(count)):
            self._key_press("\x08")

    def send_string(self, string):
        self._refresh_keyboard_layout()
        for char in self.with_delay(string):
            if char in self.keyboard_layout.char_to_vk_ss:
                # We know how to simulate the character.
                self._key_press(char)
            else:
                # Otherwise, we send it as a Unicode string.
                self._key_unicode(char)

    def send_key_combination(self, combo):
        # Make sure keyboard layout is up-to-date.
        self._refresh_keyboard_layout()
        # Parse and validate combo.
        key_events = parse_key_combo(combo, self.keyboard_layout.keyname_to_vk.get)
        # Send events...
        for keycode, pressed in self.with_delay(key_events):
            self._key_event(keycode, pressed)



import pytest

from plover.machine.keymap import Keymap


def new_keymap():
    return Keymap(("k%u" % n for n in range(8)), ("a%u" % n for n in range(4)))


BINDINGS_LIST = (
    ("k0", "a0"),
    ("k1", "a3"),
    ("k4", "a1"),
    ("k5", "a2"),
    ("k7", "a1"),
)
BINDINGS_DICT = dict(BINDINGS_LIST)
MAPPINGS_LIST = (
    ("a0", ("k0",)),
    ("a1", ("k4", "k7")),
    ("a2", ("k5",)),
    ("a3", ("k1",)),
)
MAPPINGS_DICT = dict(MAPPINGS_LIST)
MAPPINGS_FULL = dict(MAPPINGS_DICT)
MAPPINGS_FULL.update({"no-op": ()})


def test_keymap_init():
    k = new_keymap()
    assert k.get_bindings() == {}
    assert k.get_mappings() == {}


def test_keymap_set_bindings():
    # Set bindings from a dictionary.
    k = new_keymap()
    k.set_bindings(BINDINGS_DICT)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL
    # Set bindings from a list of tuples.
    k = new_keymap()
    k.set_bindings(BINDINGS_LIST)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL


def test_keymap_set_mappings():
    # Set mappings from a dictionary.
    k = new_keymap()
    k.set_mappings(MAPPINGS_DICT)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL
    # Set mappings from a list of tuples.
    k = new_keymap()
    k.set_mappings(MAPPINGS_LIST)
    assert k.get_bindings() == BINDINGS_DICT
    assert k.get_mappings() == MAPPINGS_FULL


def test_keymap_setitem():
    bindings = dict(BINDINGS_DICT)
    mappings = dict(MAPPINGS_FULL)
    k = new_keymap()
    k.set_mappings(mappings)
    # Bind to one key.
    k["a3"] = "k6"
    del bindings["k1"]
    bindings["k6"] = "a3"
    mappings["a3"] = ("k6",)
    assert k.get_bindings() == bindings
    assert k.get_mappings() == mappings
    # Bind to multiple keys.
    k["a3"] = ("k6", "k1")
    bindings["k1"] = "a3"
    bindings["k6"] = "a3"
    mappings["a3"] = (
        "k1",
        "k6",
    )
    assert k.get_bindings() == bindings
    assert k.get_mappings() == mappings
    # If a key is already mapped (k0), don't override it.
    k["a3"] = ("k0", "k2")
    del bindings["k1"]
    del bindings["k6"]
    bindings["k2"] = "a3"
    mappings["a3"] = ("k2",)
    assert k.get_bindings() == bindings
    assert k.get_mappings() == mappings
    # Assert on invalid action.
    with pytest.raises(AssertionError):
        k["a9"] = "k0"



import os
import sys
import subprocess

from plover.oslayer.config import CONFIG_DIR, PLATFORM, PLUGINS_PLATFORM


def main():
    args = sys.argv[:]
    args[0:1] = [sys.executable, "-m", "plover.scripts.main", "--gui", "qt"]
    if "--no-user-plugins" in args[3:]:
        args.remove("--no-user-plugins")
        args.insert(1, "-s")
    os.environ["PYTHONUSERBASE"] = os.path.join(CONFIG_DIR, "plugins", PLUGINS_PLATFORM)
    if PLATFORM == "win":
        # Workaround https://bugs.python.org/issue19066
        subprocess.Popen(args, cwd=os.getcwd())
        sys.exit(0)
    os.execv(args[0], args)


if __name__ == "__main__":
    main()



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for steno.py."""

import inspect

import pytest

from plover.steno import normalize_steno, Stroke

from plover_build_utils.testing import parametrize


NORMALIZE_TESTS = (
    lambda: ("S", ("S",)),
    lambda: ("S-", ("S",)),
    lambda: ("-S", ("-S",)),
    lambda: ("ES", ("ES",)),
    lambda: ("-ES", ("ES",)),
    lambda: ("TW-EPBL", ("TWEPBL",)),
    lambda: ("TWEPBL", ("TWEPBL",)),
    lambda: ("RR", ("R-R",)),
    lambda: ("19", ("1-9",)),
    lambda: ("14", ("14",)),
    lambda: ("146", ("14-6",)),
    lambda: ("67", ("-67",)),
    lambda: ("120-7", ("1207",)),
    lambda: ("6", ("-6",)),
    lambda: ("9", ("-9",)),
    lambda: ("5", ("5",)),
    lambda: ("0", ("0",)),
    lambda: ("456", ("456",)),
    lambda: ("46", ("4-6",)),
    lambda: ("4*6", ("4*6",)),
    lambda: ("456", ("456",)),
    lambda: ("S46", ("14-6",)),
    lambda: ("T-EFT/-G", ("TEFT", "-G")),
    lambda: ("T-EFT/G", ("TEFT", "-G")),
    lambda: ("/PRE", ("", "PRE")),
    lambda: ("S--T", ("S-T",)),
    # Number key.
    lambda: ("#", ("#",)),
    lambda: ("#S", ("1",)),
    lambda: ("#A", ("5",)),
    lambda: ("#0", ("0",)),
    lambda: ("#6", ("-6",)),
    # RTF/CRE spec allow the number key letter anywhere…
    lambda: ("#WS", ("#W-S",)),
    lambda: ("W#S", ("#W-S",)),
    lambda: ("W-S#", ("#W-S",)),
    lambda: ("S#A", ("15",)),
    lambda: ("2#", ("2",)),
    # Implicit hyphens.
    lambda: ("SA-", ("SA",)),
    lambda: ("SA-R", ("SAR",)),
    lambda: ("O", ("O",)),
    lambda: ("O-", ("O",)),
    lambda: ("S*-R", ("S*R",)),
    # Invalid.
    lambda: ("SRALD/invalid", (ValueError, ("SRALD", "invalid"))),
    lambda: ("SRALD//invalid", (ValueError, ("SRALD", "", "invalid"))),
    lambda: ("S-*R", (ValueError, ("S-*R",))),
    lambda: ("-O-", (ValueError, ("-O-",))),
    lambda: ("-O", (ValueError, ("-O",))),
    lambda: ("#WS#", (ValueError, ("#WS#",))),
)


@parametrize(NORMALIZE_TESTS)
@pytest.mark.parametrize("mode", ("strict=False", "strict=True"))
def test_normalize_steno(mode, steno, expected):
    kwargs = eval("dict(" + mode + ")")
    if inspect.isclass(expected[0]):
        if kwargs["strict"]:
            with pytest.raises(expected[0]):
                normalize_steno(steno)
            return
        expected = expected[1]
    result = normalize_steno(steno, **kwargs)
    msg = "normalize_steno(%r, %s)=%r != %r" % (steno, mode, result, expected)
    assert result == expected, msg


STROKE_TESTS = (
    lambda: (["S-"], (["S-"], "S")),
    lambda: (["S-", "T-"], (["S-", "T-"], "ST")),
    lambda: (["T-", "S-"], (["S-", "T-"], "ST")),
    lambda: (["-P", "-P"], (["-P"], "-P")),
    lambda: (["#", "S-", "-T"], (["#", "S-", "-T"], "1-9")),
    lambda: (["1-", "-9"], (["#", "S-", "-T"], "1-9")),
    lambda: (["-P", "X-"], ValueError),
)


@parametrize(STROKE_TESTS)
def test_stroke(keys, expected):
    if inspect.isclass(expected):
        with pytest.raises(expected):
            Stroke(keys)
    else:
        steno_keys, rtfcre = expected
        stroke = Stroke(keys)
        assert stroke.steno_keys == steno_keys
        assert stroke.rtfcre == rtfcre



from PySide6.QtGui import QValidator

import pytest

from plover.gui_qt.steno_validator import StenoValidator


@pytest.mark.parametrize(
    ("text", "state"),
    (
        # Acceptable.
        ("ST", QValidator.State.Acceptable),
        ("TEFT", QValidator.State.Acceptable),
        ("TEFT/-G", QValidator.State.Acceptable),
        ("/ST", QValidator.State.Acceptable),
        ("-F", QValidator.State.Acceptable),
        # Intermediate.
        ("-", QValidator.State.Intermediate),
        ("/", QValidator.State.Intermediate),
        ("/-", QValidator.State.Intermediate),
        ("ST/", QValidator.State.Intermediate),
        ("ST/-", QValidator.State.Intermediate),
        ("ST//", QValidator.State.Intermediate),
        # Invalid.
        ("WK", QValidator.State.Invalid),
        ("PLOVER", QValidator.State.Invalid),
    ),
)
def test_steno_validator_validate(text, state):
    validator = StenoValidator()
    assert validator.validate(text, len(text)) == (state, text, len(text))
